import os
from pathlib import Path

import pandas as pd
import pytest

from ludwig.constants import CHECKSUM, META, TEST, TRAINING, VALIDATION
from ludwig.data.cache.manager import alphanum, CacheManager
from ludwig.data.cache.types import CacheableDataframe, wrap
from ludwig.data.dataset.pandas import PandasDatasetManager
from ludwig.globals import TRAINING_PREPROC_FILE_NAME
from tests.integration_tests.utils import category_feature, LocalTestBackend, sequence_feature


@pytest.fixture
def change_test_dir(tmpdir, monkeypatch):
    monkeypatch.chdir(tmpdir)


@pytest.mark.parametrize("use_df", [True, False], ids=["df", "filename"])
@pytest.mark.parametrize("use_split", [True, False], ids=["split", "no_split"])
@pytest.mark.parametrize("use_cache_dir", [True, False], ids=["cache_dir", "no_cache_dir"])
def test_cache_dataset(use_cache_dir, use_split, use_df, tmpdir, change_test_dir):
    dataset_manager = PandasDatasetManager(backend=LocalTestBackend())
    cache_dir = os.path.join(tmpdir, "cache") if use_cache_dir else None
    manager = CacheManager(dataset_manager, cache_dir=cache_dir)

    config = {
        "input_features": [sequence_feature(encoder={"reduce_output": "sum"})],
        "output_features": [category_feature(decoder={"vocab_size": 2}, reduce_input="sum")],
        "combiner": {"type": "concat", "output_size": 14},
        "preprocessing": {},
    }

    def touch(basename):
        path = os.path.join(tmpdir, f"{basename}.csv")
        Path(path).touch()
        return path

    def create_dataset(name):
        if use_df:
            return CacheableDataframe(df=pd.DataFrame(), name=name, checksum=name)
        else:
            return wrap(touch(name))

    dataset = training_set = test_set = validation_set = None
    if not use_split:
        dataset = create_dataset("dataset")
        cache_key = manager.get_cache_key(dataset, config)
    else:
        training_set = create_dataset("train")
        test_set = create_dataset("test")
        validation_set = create_dataset("validation")
        cache_key = manager.get_cache_key(training_set, config)

    training_set_metadata = {
        CHECKSUM: cache_key,
    }

    cache = manager.get_dataset_cache(config, dataset, training_set, test_set, validation_set)
    cache_map = cache.cache_map
    assert len(cache_map) == 4

    train_path = os.path.join(cache_dir, alphanum(cache_key)) if use_cache_dir else os.path.join(tmpdir, "dataset")
    test_path = val_path = train_path

    if use_split and not use_cache_dir:
        train_path = os.path.join(tmpdir, "train")
        test_path = os.path.join(tmpdir, "test")
        val_path = os.path.join(tmpdir, "validation")

    assert cache_map[META] == f"{train_path}.meta.json"
    assert cache_map[TRAINING] == f"{train_path}.{TRAINING_PREPROC_FILE_NAME}"
    assert cache_map[TEST] == f"{test_path}.test.hdf5"
    assert cache_map[VALIDATION] == f"{val_path}.validation.hdf5"

    for cache_path in cache_map.values():
        assert not os.path.exists(cache_path)

    training_set = pd.DataFrame()
    test_set = pd.DataFrame()
    validation_set = pd.DataFrame()

    if use_cache_dir:
        os.makedirs(cache_dir)
    cache.put(training_set, test_set, validation_set, training_set_metadata)

    for cache_path in cache_map.values():
        assert os.path.exists(cache_path)

    cache.delete()

    for cache_path in cache_map.values():
        assert not os.path.exists(cache_path)



from typing import List

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import BAG
from ludwig.schema import utils as schema_utils
from ludwig.schema.encoders.base import BaseEncoderConfig
from ludwig.schema.encoders.utils import register_encoder_config
from ludwig.schema.metadata import ENCODER_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@register_encoder_config("embed", BAG)
@ludwig_dataclass
class BagEmbedWeightedConfig(BaseEncoderConfig):
    @staticmethod
    def module_name():
        return "BagEmbedWeighted"

    type: str = schema_utils.ProtectedString(
        "embed",
        description=ENCODER_METADATA["BagEmbedWeighted"]["type"].long_description,
    )

    dropout: float = schema_utils.FloatRange(
        default=0.0,
        min=0,
        max=1,
        description="Dropout probability for the embedding.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["dropout"],
    )

    activation: str = schema_utils.ActivationOptions(
        description="The default activation function that will be used for each layer.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["activation"],
    )

    vocab: List[str] = schema_utils.List(
        default=None,
        description="Vocabulary of the encoder",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["vocab"],
    )

    representation: str = schema_utils.StringOptions(
        ["dense", "sparse"],
        default="dense",
        description="The representation of the embedding. Either dense or sparse.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["representation"],
    )

    embedding_size: int = schema_utils.PositiveInteger(
        default=50,
        description="The maximum embedding size, the actual size will be min(vocabulary_size, embedding_size) for "
        "dense representations and exactly vocabulary_size for the sparse encoding, where vocabulary_size "
        "is the number of different strings appearing in the training set in the input column (plus 1 for "
        "the unknown token placeholder <UNK>).",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["embedding_size"],
    )

    force_embedding_size: bool = schema_utils.Boolean(
        default=False,
        description="Force the embedding size to be equal to the vocabulary size. This parameter has effect only if "
        "representation is dense.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["force_embedding_size"],
    )

    embeddings_on_cpu: bool = schema_utils.Boolean(
        default=False,
        description="By default embedding matrices are stored on GPU memory if a GPU is used, as it allows for faster "
        "access, but in some cases the embedding matrix may be too large. This parameter forces the "
        "placement of the embedding matrix in regular memory and the CPU is used for embedding lookup, "
        "slightly slowing down the process as a result of data transfer between CPU and GPU memory.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["embeddings_on_cpu"],
    )

    embeddings_trainable: bool = schema_utils.Boolean(
        default=True,
        description="If true embeddings are trained during the training process, if false embeddings are fixed. It "
        "may be useful when loading pretrained embeddings for avoiding fine tuning them. This parameter "
        "has effect only when representation is dense as sparse one-hot encodings are not trainable.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["embeddings_trainable"],
    )

    pretrained_embeddings: str = schema_utils.String(
        default=None,
        allow_none=True,
        description="By default dense embeddings are initialized randomly, but this parameter allows to specify a "
        "path to a file containing embeddings in the GloVe format. When the file containing the "
        "embeddings is loaded, only the embeddings with labels present in the vocabulary are kept, "
        "the others are discarded. If the vocabulary contains strings that have no match in the "
        "embeddings file, their embeddings are initialized with the average of all other embedding plus "
        "some random noise to make them different from each other. This parameter has effect only if "
        "representation is dense.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["pretrained_embeddings"],
    )

    use_bias: bool = schema_utils.Boolean(
        default=True,
        description="Whether the layer uses a bias vector.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["use_bias"],
    )

    bias_initializer: str = schema_utils.InitializerOptions(
        default="zeros",
        description="Initializer to use for the bias vector.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["bias_initializer"],
    )

    weights_initializer: str = schema_utils.InitializerOptions(
        description="Initializer to use for the weights matrix.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["weights_initializer"],
    )

    output_size: int = schema_utils.PositiveInteger(
        default=10,
        description="If output_size is not already specified in fc_layers this is the default output_size that will "
        "be used for each layer. It indicates the size of the output of a fully connected layer.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["output_size"],
    )

    norm: str = schema_utils.StringOptions(
        ["batch", "layer"],
        default=None,
        allow_none=True,
        description="The default norm that will be used for each layer.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["norm"],
    )

    norm_params: dict = schema_utils.Dict(
        default=None,
        description="Parameters used if norm is either `batch` or `layer`.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["norm_params"],
    )

    num_fc_layers: int = schema_utils.NonNegativeInteger(
        default=0,
        description="This is the number of stacked fully connected layers that the input to the feature passes "
        "through. Their output is projected in the feature's output space.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["num_fc_layers"],
    )

    fc_layers: List[dict] = schema_utils.DictList(  # TODO (Connor): Add nesting logic for fc_layers
        default=None,
        description="List of dictionaries containing the parameters for each fully connected layer.",
        parameter_metadata=ENCODER_METADATA["BagEmbedWeighted"]["fc_layers"],
    )



import os
from typing import Any, Dict, Union

import yaml

from ludwig.schema.metadata.parameter_metadata import ParameterMetadata

_PATH_HERE = os.path.abspath(os.path.dirname(__file__))
_CONFIG_DIR = os.path.join(_PATH_HERE, "configs")


def _to_metadata(d: Dict[str, Any]) -> Union[ParameterMetadata, Dict[str, Any]]:
    is_nested = False
    for k, v in list(d.items()):
        if isinstance(v, dict):
            d[k] = _to_metadata(v)
            is_nested = True

    if is_nested:
        return d

    return ParameterMetadata.from_dict(d)


def _load(fname: str) -> Dict[str, Any]:
    with open(os.path.join(_CONFIG_DIR, fname)) as f:
        return _to_metadata(yaml.safe_load(f))


COMMON_METADATA = _load("common.yaml")
COMBINER_METADATA = _load("combiners.yaml")
DECODER_METADATA = _load("decoders.yaml")
ENCODER_METADATA = _load("encoders.yaml")
FEATURE_METADATA = _load("features.yaml")
PREPROCESSING_METADATA = _load("preprocessing.yaml")
TRAINER_METADATA = _load("trainer.yaml")
OPTIMIZER_METADATA = _load("optimizers.yaml")
LOSS_METADATA = _load("loss.yaml")



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import Any, Callable, Generator, Optional, Type

import torch
from torch import Tensor, tensor
from torchmetrics import AUROC, CharErrorRate, MeanAbsoluteError
from torchmetrics import MeanMetric as _MeanMetric
from torchmetrics import MeanSquaredError, Metric
from torchmetrics.classification import (
    BinaryAccuracy,
    BinaryPrecision,
    BinaryRecall,
    BinarySpecificity,
    MulticlassAccuracy,
)
from torchmetrics.functional.regression.r2 import _r2_score_compute, _r2_score_update
from torchmetrics.metric import jit_distributed_available
from torchmetrics.text.perplexity import Perplexity

from ludwig.constants import (
    ACCURACY,
    BINARY,
    BINARY_WEIGHTED_CROSS_ENTROPY,
    CATEGORY,
    HITS_AT_K,
    JACCARD,
    LOGITS,
    LOSS,
    MAXIMIZE,
    MEAN_ABSOLUTE_ERROR,
    MEAN_SQUARED_ERROR,
    MINIMIZE,
    NUMBER,
    PERPLEXITY,
    PRECISION,
    PREDICTIONS,
    PROBABILITIES,
    R2,
    RECALL,
    ROC_AUC,
    ROOT_MEAN_SQUARED_ERROR,
    ROOT_MEAN_SQUARED_PERCENTAGE_ERROR,
    SEQUENCE,
    SEQUENCE_ACCURACY,
    SET,
    SPECIFICITY,
    TEXT,
    TOKEN_ACCURACY,
    VECTOR,
)
from ludwig.distributed import get_current_dist_strategy
from ludwig.modules.loss_modules import (
    BWCEWLoss,
    SequenceSoftmaxCrossEntropyLoss,
    SigmoidCrossEntropyLoss,
    SoftmaxCrossEntropyLoss,
)
from ludwig.modules.metric_registry import get_metric_objective, get_metric_registry, register_metric
from ludwig.utils.loss_utils import rmspe_loss
from ludwig.utils.metric_utils import masked_correct_predictions
from ludwig.utils.torch_utils import sequence_length_2D

logger = logging.getLogger(__name__)


def _gather_all_tensors_fn() -> Optional[Callable]:
    get_current_dist_strategy().gather_all_tensors_fn()


class LudwigMetric(Metric, ABC):
    @classmethod
    def can_report(cls, feature: "OutputFeature") -> bool:  # noqa: F821
        return True

    @contextmanager
    def sync_context(
        self,
        dist_sync_fn: Optional[Callable] = None,
        process_group: Optional[Any] = None,
        should_sync: bool = True,
        should_unsync: bool = True,
        distributed_available: Optional[Callable] = jit_distributed_available,
    ) -> Generator:
        """Override the behavior of this in the base class to support custom distributed strategies."""
        dist_strategy = get_current_dist_strategy()
        self.sync(
            dist_sync_fn=dist_strategy.gather_all_tensors_fn(),
            process_group=process_group,
            should_sync=should_sync,
            distributed_available=dist_strategy.is_available,
        )

        yield

        self.unsync(should_unsync=self._is_synced and should_unsync)


@register_metric(ROOT_MEAN_SQUARED_ERROR, [NUMBER], MINIMIZE, PREDICTIONS)
class RMSEMetric(MeanSquaredError, LudwigMetric):
    """Root mean squared error metric."""

    def __init__(self, **kwargs):
        super().__init__(squared=False, dist_sync_fn=_gather_all_tensors_fn(), **kwargs)


@register_metric(PRECISION, [BINARY], MAXIMIZE, PROBABILITIES)
class PrecisionMetric(BinaryPrecision, LudwigMetric):
    """Precision metric."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())


@register_metric(RECALL, [BINARY], MAXIMIZE, PROBABILITIES)
class RecallMetric(BinaryRecall, LudwigMetric):
    """Recall metric."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())


@register_metric(ROC_AUC, [BINARY], MAXIMIZE, PROBABILITIES)
class BinaryAUROCMetric(AUROC, LudwigMetric):
    """Area under the receiver operating curve."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target.type(torch.int8))


@register_metric(ROC_AUC, [CATEGORY], MAXIMIZE, PROBABILITIES)
class CategoryAUROCMetric(AUROC, LudwigMetric):
    """Area under the receiver operating curve."""

    def __init__(self, num_classes: int, **kwargs):
        super().__init__(num_classes=num_classes, dist_sync_fn=_gather_all_tensors_fn())


@register_metric(SPECIFICITY, [BINARY], MAXIMIZE, PROBABILITIES)
class SpecificityMetric(BinarySpecificity, LudwigMetric):
    """Specificity metric."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())


class MeanMetric(LudwigMetric):
    """Abstract class for computing mean of metrics."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())
        self.avg = _MeanMetric()

    def update(self, preds: Tensor, target: Tensor) -> None:
        self.avg.update(self.get_current_value(preds, target))

    def compute(self) -> Tensor:
        return self.avg.compute()

    def reset(self):
        super().reset()
        self.avg.reset()

    @abstractmethod
    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        raise NotImplementedError()


@register_metric(ROOT_MEAN_SQUARED_PERCENTAGE_ERROR, [NUMBER], MINIMIZE, PREDICTIONS)
class RMSPEMetric(MeanMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    """ Root mean squared percentage error metric. """

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return rmspe_loss(target, preds)


@register_metric(R2, [NUMBER, VECTOR], MAXIMIZE, PREDICTIONS)
class R2Score(LudwigMetric):
    """Custom R-squared metric implementation that modifies torchmetrics R-squared implementation to return Nan
    when there is only sample. This is because R-squared is only defined for two or more samples.

    Custom implementation uses code from torchmetrics v0.9.2's implementation of R2: https://github.com/Lightning-
    AI/metrics/blob/master/src/torchmetrics/regression/r2.py
    """

    def __init__(
        self, num_outputs: int = 1, adjusted: int = 0, multioutput: str = "uniform_average", **kwargs: Any
    ) -> None:
        super().__init__(**kwargs)

        self.num_outputs = num_outputs

        if adjusted < 0 or not isinstance(adjusted, int):
            raise ValueError("`adjusted` parameter should be an integer larger or equal to 0.")
        self.adjusted = adjusted

        allowed_multioutput = ("raw_values", "uniform_average", "variance_weighted")
        if multioutput not in allowed_multioutput:
            raise ValueError(
                f"Invalid input to argument `multioutput`. Choose one of the following: {allowed_multioutput}"
            )
        self.multioutput = multioutput

        self.add_state("sum_squared_error", default=torch.zeros(self.num_outputs), dist_reduce_fx="sum")
        self.add_state("sum_error", default=torch.zeros(self.num_outputs), dist_reduce_fx="sum")
        self.add_state("residual", default=torch.zeros(self.num_outputs), dist_reduce_fx="sum")
        self.add_state("total", default=tensor(0), dist_reduce_fx="sum")

    def update(self, preds: Tensor, target: Tensor) -> None:
        """Update state with predictions and targets.

        Args:
            preds: Predictions from model
            target: Ground truth values
        """
        sum_squared_error, sum_error, residual, n_obs = _r2_score_update(preds, target)

        self.sum_squared_error += sum_squared_error
        self.sum_error += sum_error
        self.residual += residual
        self.total += n_obs

    def compute(self) -> Tensor:
        """Computes r2 score over the metric states."""

        # self.total maps to the number of observations in preds/target computed during update()
        if self.total <= 1:
            logger.warning(
                """R-squared (r2) is not defined for one sample. It needs at least two samples. Returning NaN."""
            )
            return torch.tensor(float("nan"))

        return _r2_score_compute(
            self.sum_squared_error, self.sum_error, self.residual, self.total, self.adjusted, self.multioutput
        )


@register_metric(LOSS, [], MINIMIZE, LOGITS)
class LossMetric(MeanMetric, ABC):
    def __init__(self):
        super().__init__()

    @abstractmethod
    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        raise NotImplementedError()

    @classmethod
    def can_report(cls, feature: "OutputFeature") -> bool:  # noqa: F821
        return False


@register_metric(BINARY_WEIGHTED_CROSS_ENTROPY, [BINARY], MINIMIZE, LOGITS)
class BWCEWLMetric(LossMetric):
    """Binary Weighted Cross Entropy Weighted Logits Score Metric."""

    def __init__(
        self,
        positive_class_weight: Optional[Tensor] = None,
        robust_lambda: int = 0,
        confidence_penalty: int = 0,
        **kwargs,
    ):
        super().__init__()

        self.loss_function = BWCEWLoss(
            positive_class_weight=positive_class_weight,
            robust_lambda=robust_lambda,
            confidence_penalty=confidence_penalty,
        )

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return self.loss_function(preds, target)


@register_metric("softmax_cross_entropy", [CATEGORY], MINIMIZE, LOGITS)
class SoftmaxCrossEntropyMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.softmax_cross_entropy_function = SoftmaxCrossEntropyLoss(**kwargs)

    def get_current_value(self, preds: Tensor, target: Tensor):
        return self.softmax_cross_entropy_function(preds, target)


@register_metric("sequence_softmax_cross_entropy", [SEQUENCE, TEXT], MINIMIZE, LOGITS)
class SequenceSoftmaxCrossEntropyMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.sequence_softmax_cross_entropy_function = SequenceSoftmaxCrossEntropyLoss(**kwargs)

    def get_current_value(self, preds: Tensor, target: Tensor):
        return self.sequence_softmax_cross_entropy_function(preds, target)


@register_metric("sigmoid_cross_entropy", [SET], MINIMIZE, LOGITS)
class SigmoidCrossEntropyMetric(LossMetric):
    def __init__(self, **kwargs):
        super().__init__()
        self.sigmoid_cross_entropy_function = SigmoidCrossEntropyLoss(**kwargs)

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return self.sigmoid_cross_entropy_function(preds, target)


@register_metric(TOKEN_ACCURACY, [SEQUENCE, TEXT], MAXIMIZE, PREDICTIONS)
class TokenAccuracyMetric(MeanMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        target = target.type(preds.dtype)
        target_sequence_length = sequence_length_2D(target)
        masked_correct_preds = masked_correct_predictions(target, preds, target_sequence_length)
        return torch.mean(masked_correct_preds)


@register_metric(SEQUENCE_ACCURACY, [SEQUENCE, TEXT], MAXIMIZE, PREDICTIONS)
class SequenceAccuracyMetric(MeanMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        return torch.sum(torch.all(preds == target, dim=1)) / target.size()[0]


@register_metric(PERPLEXITY, [SEQUENCE, TEXT], MINIMIZE, PROBABILITIES)
class PerplexityMetric(Perplexity, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target.type(torch.int64))


@register_metric("char_error_rate", [SEQUENCE, TEXT], MINIMIZE, PREDICTIONS)
class CharErrorRateMetric(CharErrorRate, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())


@register_metric(ACCURACY, [BINARY], MAXIMIZE, PREDICTIONS)
class Accuracy(BinaryAccuracy, LudwigMetric):
    """R-squared metric."""

    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())


@register_metric(ACCURACY, [CATEGORY], MAXIMIZE, PREDICTIONS)
class CategoryAccuracy(MulticlassAccuracy, LudwigMetric):
    def __init__(self, num_classes: int, **kwargs):
        super().__init__(num_classes=num_classes, dist_sync_fn=_gather_all_tensors_fn())

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target.type(torch.long))


@register_metric(HITS_AT_K, [CATEGORY], MAXIMIZE, LOGITS)
class HitsAtKMetric(MulticlassAccuracy, LudwigMetric):
    def __init__(self, num_classes: int, top_k: int, **kwargs):
        super().__init__(num_classes=num_classes, top_k=top_k, dist_sync_fn=_gather_all_tensors_fn(), **kwargs)

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target.type(torch.long))

    @classmethod
    def can_report(cls, feature: "OutputFeature") -> bool:  # noqa: F821
        return feature.num_classes > feature.top_k


@register_metric(MEAN_ABSOLUTE_ERROR, [NUMBER, VECTOR], MINIMIZE, PREDICTIONS)
class MAEMetric(MeanAbsoluteError, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds.detach(), target)


@register_metric(MEAN_SQUARED_ERROR, [NUMBER, VECTOR], MINIMIZE, PREDICTIONS)
class MSEMetric(MeanSquaredError, LudwigMetric):
    def __init__(self, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())

    def update(self, preds: Tensor, target: Tensor) -> None:
        super().update(preds, target)


@register_metric(JACCARD, [SET], MAXIMIZE, PROBABILITIES)
class JaccardMetric(MeanMetric):
    def __init__(self, threshold: float = 0.5, **kwargs):
        super().__init__(dist_sync_fn=_gather_all_tensors_fn())
        self.threshold = threshold

    def get_current_value(self, preds: Tensor, target: Tensor) -> Tensor:
        # notation: b is batch size and nc is number of unique elements in the set
        # preds: shape [b, nc] probabilities for each class
        # target: shape [b, nc] bit-mapped set representation
        preds = torch.greater_equal(preds, self.threshold)  # now bit-mapped set
        target = target.type(torch.bool)

        intersection = torch.sum(torch.logical_and(target, preds).type(torch.float32), dim=-1)
        union = torch.sum(torch.logical_or(target, preds).type(torch.float32), dim=-1)

        return intersection / union  # shape [b]


def get_metric_cls(metric_name: str) -> Type[LudwigMetric]:
    return get_metric_registry()[metric_name]


def get_improved_fn(metric: str) -> Callable:
    if get_metric_objective(metric) == MINIMIZE:
        return lambda x, y: x < y
    else:
        return lambda x, y: x > y


def get_initial_validation_value(metric: str) -> float:
    if get_metric_objective(metric) == MINIMIZE:
        return float("inf")
    else:
        return float("-inf")


def get_best_function(metric: str) -> Callable:
    if get_metric_objective(metric) == MINIMIZE:
        return min
    else:
        return max



#! /usr/bin/env python
# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from typing import Any, Dict, List, Union

import numpy as np
import torch

from ludwig.constants import (
    CATEGORY,
    COLUMN,
    HIDDEN,
    LOGITS,
    NAME,
    PREDICTIONS,
    PROBABILITIES,
    PROBABILITY,
    PROC_COLUMN,
    PROJECTION_INPUT,
)
from ludwig.error import InputDataError
from ludwig.features.base_feature import BaseFeatureMixin, InputFeature, OutputFeature, PredictModule
from ludwig.schema.features.category_feature import CategoryInputFeatureConfig, CategoryOutputFeatureConfig
from ludwig.types import (
    FeatureMetadataDict,
    FeaturePostProcessingOutputDict,
    PreprocessingConfigDict,
    TrainingSetMetadataDict,
)
from ludwig.utils import calibration, output_feature_utils
from ludwig.utils.eval_utils import ConfusionMatrix
from ludwig.utils.math_utils import int_type, softmax
from ludwig.utils.strings_utils import create_vocabulary_single_token, UNKNOWN_SYMBOL
from ludwig.utils.types import TorchscriptPreprocessingInput

logger = logging.getLogger(__name__)


class _CategoryPreprocessing(torch.nn.Module):
    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.str2idx = metadata["str2idx"]
        if UNKNOWN_SYMBOL in self.str2idx:
            self.unk = self.str2idx[UNKNOWN_SYMBOL]
        else:
            # self.unk is set to 0 to comply with Torchscript type tracing and will
            # likely not be used during training, but potentially during inference
            self.unk = 0

    def forward(self, v: TorchscriptPreprocessingInput) -> torch.Tensor:
        if not torch.jit.isinstance(v, List[str]):
            raise ValueError(f"Unsupported input: {v}")

        indices = [self.str2idx.get(s.strip(), self.unk) for s in v]
        return torch.tensor(indices, dtype=torch.int32)


class _CategoryPostprocessing(torch.nn.Module):
    def __init__(self, metadata: TrainingSetMetadataDict):
        super().__init__()
        self.idx2str = {i: v for i, v in enumerate(metadata["idx2str"])}
        self.unk = UNKNOWN_SYMBOL
        self.predictions_key = PREDICTIONS
        self.probabilities_key = PROBABILITIES

    def forward(self, preds: Dict[str, torch.Tensor], feature_name: str) -> FeaturePostProcessingOutputDict:
        predictions = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.predictions_key)
        probabilities = output_feature_utils.get_output_feature_tensor(preds, feature_name, self.probabilities_key)

        inv_preds = [self.idx2str.get(pred, self.unk) for pred in predictions]

        return {
            self.predictions_key: inv_preds,
            self.probabilities_key: probabilities,
        }


class _CategoryPredict(PredictModule):
    def __init__(self, calibration_module=None):
        super().__init__()
        self.calibration_module = calibration_module

    def forward(self, inputs: Dict[str, torch.Tensor], feature_name: str) -> Dict[str, torch.Tensor]:
        logits = output_feature_utils.get_output_feature_tensor(inputs, feature_name, self.logits_key)

        if self.calibration_module is not None:
            probabilities = self.calibration_module(logits)
        else:
            probabilities = torch.softmax(logits, -1)

        predictions = torch.argmax(probabilities, -1)
        predictions = predictions.long()

        # EXPECTED SHAPE OF RETURNED TENSORS
        # predictions: [batch_size]
        # probabilities: [batch_size, num_classes]
        # logits: [batch_size, num_classes]
        return {self.predictions_key: predictions, self.probabilities_key: probabilities, self.logits_key: logits}


class CategoryFeatureMixin(BaseFeatureMixin):
    @staticmethod
    def type():
        return CATEGORY

    @staticmethod
    def cast_column(column, backend):
        return column.astype(str)

    @staticmethod
    def get_feature_meta(
        column, preprocessing_parameters: PreprocessingConfigDict, backend, is_input_feature: bool
    ) -> FeatureMetadataDict:
        idx2str, str2idx, str2freq = create_vocabulary_single_token(
            column,
            num_most_frequent=preprocessing_parameters["most_common"],
            processor=backend.df_engine,
        )
        vocab_size = len(str2idx)
        if not is_input_feature and vocab_size <= 1:
            # Category output feature with vocab size 1
            raise InputDataError(
                column.name,
                CATEGORY,
                f"""
                At least 2 distinct values are required for category output features, but column
                only contains {str(idx2str)}.
                """,
            )
        if vocab_size <= 1:
            # Category input feature with vocab size 1
            logger.info(
                f"Input feature '{column.name}' contains only 1 distinct value {str(idx2str)}. This is not useful"
                " for machine learning models because this feature has zero variance. Consider removing this feature"
                " from your input features."
            )
        return {"idx2str": idx2str, "str2idx": str2idx, "str2freq": str2freq, "vocab_size": vocab_size}

    @staticmethod
    def feature_data(backend, column, metadata):
        def __replace_token_with_idx(value: Any, metadata: TrainingSetMetadataDict, fallback_symbol_idx: int) -> int:
            stripped_value = value.strip()
            if stripped_value in metadata["str2idx"]:
                return metadata["str2idx"][stripped_value]
            logger.warning(
                f"""
                Encountered unknown symbol '{stripped_value}' for '{column.name}' during category
                feature preprocessing. This should never happen during training. If this happens during
                inference, this may be an indication that not all possible symbols were present in your
                training set. Consider re-splitting your data to ensure full representation, or setting
                preprocessing.most_common parameter to be smaller than this feature's total vocabulary
                size, {len(metadata["str2idx"])}, which will ensure that the model is architected and
                trained with an UNKNOWN symbol. Returning the index for the most frequent symbol,
                {metadata["idx2str"][fallback_symbol_idx]}, instead.
                """
            )
            return fallback_symbol_idx

        # No unknown symbol in Metadata from preprocessing means that all values
        # should be mappable to vocabulary
        if UNKNOWN_SYMBOL not in metadata["str2idx"]:
            # If no unknown is defined, just use the most popular token's index as the fallback index
            most_popular_token = max(metadata["str2freq"], key=metadata["str2freq"].get)
            most_popular_token_idx = metadata["str2idx"].get(most_popular_token)
            return backend.df_engine.map_objects(
                column,
                lambda x: __replace_token_with_idx(x, metadata, most_popular_token_idx),
                meta=(column.name, int),
            ).astype(int_type(metadata["vocab_size"]))
        else:
            return backend.df_engine.map_objects(
                column,
                lambda x: (
                    metadata["str2idx"][x.strip()]
                    if x.strip() in metadata["str2idx"]
                    else metadata["str2idx"][UNKNOWN_SYMBOL]
                ),
                meta=(column.name, int),
            ).astype(int_type(metadata["vocab_size"]))

    @staticmethod
    def add_feature_data(
        feature_config,
        input_df,
        proc_df,
        metadata,
        preprocessing_parameters: PreprocessingConfigDict,
        backend,
        skip_save_processed_input,
    ):
        proc_df[feature_config[PROC_COLUMN]] = CategoryFeatureMixin.feature_data(
            backend,
            input_df[feature_config[COLUMN]],
            metadata[feature_config[NAME]],
        )

        return proc_df


class CategoryInputFeature(CategoryFeatureMixin, InputFeature):
    def __init__(self, input_feature_config: CategoryInputFeatureConfig, encoder_obj=None, **kwargs):
        super().__init__(input_feature_config, **kwargs)

        if encoder_obj:
            self.encoder_obj = encoder_obj
        else:
            self.encoder_obj = self.initialize_encoder(input_feature_config.encoder)

    def forward(self, inputs):
        assert isinstance(inputs, torch.Tensor)
        assert (
            inputs.dtype == torch.int8
            or inputs.dtype == torch.int16
            or inputs.dtype == torch.int32
            or inputs.dtype == torch.int64
        )
        assert len(inputs.shape) == 1 or (len(inputs.shape) == 2 and inputs.shape[1] == 1)

        if len(inputs.shape) == 1:
            inputs = inputs.unsqueeze(dim=1)

        if inputs.dtype == torch.int8 or inputs.dtype == torch.int16:
            inputs = inputs.type(torch.int)
        encoder_output = self.encoder_obj(inputs)

        return {"encoder_output": encoder_output}

    @property
    def input_dtype(self):
        return torch.int32

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([1])

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self.encoder_obj.output_shape)

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.encoder.vocab = feature_metadata["idx2str"]

    @staticmethod
    def get_schema_cls():
        return CategoryInputFeatureConfig

    @staticmethod
    def create_preproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _CategoryPreprocessing(metadata)


class CategoryOutputFeature(CategoryFeatureMixin, OutputFeature):
    def __init__(
        self,
        output_feature_config: Union[CategoryOutputFeatureConfig, Dict],
        output_features: Dict[str, OutputFeature],
        **kwargs,
    ):
        self.num_classes = output_feature_config.num_classes
        self.top_k = output_feature_config.top_k
        super().__init__(output_feature_config, output_features, **kwargs)
        if hasattr(output_feature_config.decoder, "num_classes"):
            output_feature_config.decoder.num_classes = output_feature_config.num_classes
        self.decoder_obj = self.initialize_decoder(output_feature_config.decoder)
        self._setup_loss()
        self._setup_metrics()

    def logits(self, inputs, **kwargs):  # hidden
        hidden = inputs[HIDDEN]

        # EXPECTED SHAPES FOR RETURNED TENSORS
        # logits: shape [batch_size, num_classes]
        # hidden: shape [batch_size, size of final fully connected layer]
        return {LOGITS: self.decoder_obj(hidden), PROJECTION_INPUT: hidden}

    def create_calibration_module(self, feature: CategoryOutputFeatureConfig) -> torch.nn.Module:
        """Creates the appropriate calibration module based on the feature config.

        Today, only one type of calibration ("temperature_scaling") is available, but more options may be supported in
        the future.
        """
        if feature.calibration:
            calibration_cls = calibration.get_calibration_cls(CATEGORY, "temperature_scaling")
            return calibration_cls(num_classes=self.num_classes)
        return None

    def create_predict_module(self) -> PredictModule:
        return _CategoryPredict(calibration_module=self.calibration_module)

    def get_prediction_set(self):
        return {PREDICTIONS, PROBABILITIES, LOGITS}

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size([self.input_size])

    @classmethod
    def get_output_dtype(cls):
        return torch.int64

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size([1])

    def metric_kwargs(self):
        return {"top_k": self.top_k, "num_classes": self.num_classes, "task": "multiclass"}

    @staticmethod
    def update_config_with_metadata(feature_config, feature_metadata, *args, **kwargs):
        feature_config.num_classes = feature_metadata["vocab_size"]
        feature_config.top_k = min(feature_config.num_classes, feature_config.top_k)

        if isinstance(feature_config.loss.class_weights, (list, tuple)):
            if len(feature_config.loss.class_weights) != feature_config.num_classes:
                raise ValueError(
                    "The length of class_weights ({}) is not compatible with "
                    "the number of classes ({}) for feature {}. "
                    "Check the metadata JSON file to see the classes "
                    "and their order and consider there needs to be a weight "
                    "for the <UNK> class too.".format(
                        len(feature_config.loss.class_weights),
                        feature_config.num_classes,
                        feature_config.column,
                    )
                )

        if isinstance(feature_config.loss.class_weights, dict):
            if feature_metadata["str2idx"].keys() != feature_config.loss.class_weights.keys():
                raise ValueError(
                    "The class_weights keys ({}) are not compatible with "
                    "the classes ({}) of feature {}. "
                    "Check the metadata JSON file to see the classes "
                    "and consider there needs to be a weight "
                    "for the <UNK> class too.".format(
                        feature_config.loss.class_weights.keys(),
                        feature_metadata["str2idx"].keys(),
                        feature_config.column,
                    )
                )
            else:
                class_weights = feature_config.loss.class_weights
                idx2str = feature_metadata["idx2str"]
                class_weights_list = [class_weights[s] for s in idx2str]
                feature_config.loss.class_weights = class_weights_list

        if feature_config.loss.class_similarities_temperature > 0:
            if "class_similarities" in feature_config.loss:
                similarities = feature_config.loss.class_similarities
                temperature = feature_config.loss.class_similarities_temperature

                curr_row = 0
                first_row_length = 0
                is_first_row = True
                for row in similarities:
                    if is_first_row:
                        first_row_length = len(row)
                        is_first_row = False
                        curr_row += 1
                    else:
                        curr_row_length = len(row)
                        if curr_row_length != first_row_length:
                            raise ValueError(
                                "The length of row {} of the class_similarities "
                                "of {} is {}, different from the length of "
                                "the first row {}. All rows must have "
                                "the same length.".format(
                                    curr_row, feature_config.column, curr_row_length, first_row_length
                                )
                            )
                        else:
                            curr_row += 1
                all_rows_length = first_row_length

                if all_rows_length != len(similarities):
                    raise ValueError(
                        "The class_similarities matrix of {} has "
                        "{} rows and {} columns, "
                        "their number must be identical.".format(
                            feature_config.column, len(similarities), all_rows_length
                        )
                    )

                if all_rows_length != feature_config.num_classes:
                    raise ValueError(
                        "The size of the class_similarities matrix of {} is "
                        "{}, different from the number of classes ({}). "
                        "Check the metadata JSON file to see the classes "
                        "and their order and "
                        "consider <UNK> class too.".format(
                            feature_config.column, all_rows_length, feature_config.num_classes
                        )
                    )

                similarities = np.array(similarities, dtype=np.float32)
                for i in range(len(similarities)):
                    similarities[i, :] = softmax(similarities[i, :], temperature=temperature)

                feature_config.loss.class_similarities = similarities
            else:
                raise ValueError(
                    "class_similarities_temperature > 0, "
                    "but no class_similarities are provided "
                    "for feature {}".format(feature_config.column)
                )

    @staticmethod
    def calculate_overall_stats(predictions, targets, train_set_metadata):
        overall_stats = {}
        confusion_matrix = ConfusionMatrix(targets, predictions[PREDICTIONS], labels=train_set_metadata["idx2str"])
        overall_stats["confusion_matrix"] = confusion_matrix.cm.tolist()
        overall_stats["overall_stats"] = confusion_matrix.stats()
        overall_stats["per_class_stats"] = confusion_matrix.per_class_stats()

        return overall_stats

    def postprocess_predictions(
        self,
        predictions,
        metadata,
    ):
        predictions_col = f"{self.feature_name}_{PREDICTIONS}"
        if predictions_col in predictions:
            if "idx2str" in metadata:
                predictions[predictions_col] = predictions[predictions_col].map(lambda pred: metadata["idx2str"][pred])

        probabilities_col = f"{self.feature_name}_{PROBABILITIES}"
        if probabilities_col in predictions:
            prob_col = f"{self.feature_name}_{PROBABILITY}"
            predictions[prob_col] = predictions[probabilities_col].map(max)
            predictions[probabilities_col] = predictions[probabilities_col].map(lambda pred: pred.tolist())
            if "idx2str" in metadata:
                for i, label in enumerate(metadata["idx2str"]):
                    key = f"{probabilities_col}_{label}"

                    # Use default param to force a capture before the loop completes, see:
                    # https://stackoverflow.com/questions/2295290/what-do-lambda-function-closures-capture
                    predictions[key] = predictions[probabilities_col].map(
                        lambda prob, i=i: prob[i],
                    )

        top_k_col = f"{self.feature_name}_predictions_top_k"
        if top_k_col in predictions:
            if "idx2str" in metadata:
                predictions[top_k_col] = predictions[top_k_col].map(
                    lambda pred_top_k: [metadata["idx2str"][pred] for pred in pred_top_k]
                )

        return predictions

    @staticmethod
    def get_schema_cls():
        return CategoryOutputFeatureConfig

    @staticmethod
    def create_postproc_module(metadata: TrainingSetMetadataDict) -> torch.nn.Module:
        return _CategoryPostprocessing(metadata)



#! /usr/bin/env python
# Copyright (c) 2022 Predibase, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import copy
import logging
import warnings
from typing import Any, Callable, Dict, List, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.constants import (
    AUDIO,
    BIAS,
    CLASS_WEIGHTS,
    COLUMN,
    CONV_BIAS,
    CONV_USE_BIAS,
    DECODER,
    DEFAULT_BIAS,
    DEFAULT_USE_BIAS,
    DEFAULTS,
    ENCODER,
    EVAL_BATCH_SIZE,
    EXECUTOR,
    FORCE_SPLIT,
    HEIGHT,
    HYPEROPT,
    IMAGE,
    INPUT_FEATURES,
    LOSS,
    MISSING_VALUE_STRATEGY,
    MODEL_ECD,
    MODEL_GBM,
    MODEL_TYPE,
    NAME,
    NUM_SAMPLES,
    NUMBER,
    OUTPUT_FEATURES,
    PARAMETERS,
    PREPROCESSING,
    PROBABILITIES,
    RANDOM,
    RAY,
    SAMPLER,
    SCHEDULER,
    SEARCH_ALG,
    SEQUENCE,
    SPLIT,
    SPLIT_PROBABILITIES,
    STRATIFY,
    TEXT,
    TIMESERIES,
    TRAINER,
    TRAINING,
    TYPE,
    USE_BIAS,
    WIDTH,
)
from ludwig.features.feature_registries import get_base_type_registry, get_input_type_registry, get_output_type_registry
from ludwig.globals import LUDWIG_VERSION
from ludwig.schema.defaults.gbm import GBMDefaultsConfig
from ludwig.schema.encoders.utils import get_encoder_cls
from ludwig.types import (
    FeatureConfigDict,
    HyperoptConfigDict,
    ModelConfigDict,
    PreprocessingConfigDict,
    TrainerConfigDict,
    TrainingSetMetadataDict,
)
from ludwig.utils.metric_utils import TrainerMetric
from ludwig.utils.misc_utils import get_from_registry, merge_dict
from ludwig.utils.version_transformation import VersionTransformation, VersionTransformationRegistry

config_transformation_registry = VersionTransformationRegistry()


@DeveloperAPI
def register_config_transformation(version: str, prefixes: Union[str, List[str]] = []) -> Callable:
    """This decorator registers a transformation function for a config version. Version is the first version which
    requires the transform. For example, since "training" is renamed to "trainer" in 0.5, this change should be
    registered with 0.5.  from_version < version <= to_version.

    Args:
        version: The version to register this transformation with. The earliest ludwig version which requires this
                 transformation.
        prefixes: A list of keypath prefixes to apply this transformation to. If not specified, transforms the entire
                  config dict. If a prefix indicates a list, i.e. "input_features", the transformation is applied to
                  each element of the list (each input feature).
    """
    if isinstance(prefixes, str):
        prefixes = [prefixes]

    def wrap(fn: Callable[[Dict], Dict]):
        config_transformation_registry.register(VersionTransformation(transform=fn, version=version, prefixes=prefixes))
        return fn

    return wrap


@DeveloperAPI
def upgrade_config_dict_to_latest_version(config: ModelConfigDict) -> ModelConfigDict:
    """Updates config from an older version of Ludwig to the current version. If config does not have a
    "ludwig_version" key, all updates are applied.

    Args:
        config: A config saved by an older version of Ludwig.

    Returns A new copy of config, upgraded to the current Ludwig version. Returns config if config has no
            "ludwig_version".
    """
    return config_transformation_registry.update_config(
        config, from_version=config.get("ludwig_version", "0.0"), to_version=LUDWIG_VERSION
    )


def upgrade_model_progress(model_progress: Dict) -> Dict:
    """Updates model progress info to be compatible with latest ProgressTracker implementation.

    Notably, we convert epoch-based stats to their step-based equivalents and reformat metrics into `TrainerMetric`
    tuples.
    """
    ret = copy.deepcopy(model_progress)

    if "last_improvement_epoch" in ret:
        ret["last_improvement_steps"] = ret["last_improvement_epoch"] * ret["batch_size"]
        del ret["last_improvement_epoch"]

    if "last_learning_rate_reduction_epoch" in ret:
        ret["last_learning_rate_reduction_steps"] = ret["last_learning_rate_reduction_epoch"] * ret["batch_size"]
        del ret["last_learning_rate_reduction_epoch"]

    if "last_increase_batch_size_epoch" in ret:
        ret["last_increase_batch_size_steps"] = ret["last_increase_batch_size_epoch"] * ret["batch_size"]
        del ret["last_increase_batch_size_epoch"]

    if "vali_metrics" in ret:
        ret["validation_metrics"] = ret["vali_metrics"]
        del ret["vali_metrics"]

    for metric_group in ("train_metrics", "test_metrics", "validation_metrics"):
        for tgt in ret[metric_group]:
            for metric in ret[metric_group][tgt]:
                if len(ret[metric_group][tgt][metric]) == 0 or isinstance(
                    ret[metric_group][tgt][metric][0], (tuple, list)
                ):
                    continue

                ret[metric_group][tgt][metric] = [
                    TrainerMetric(i + 1, (i + 1) * ret["batch_size"], val)
                    for i, val in enumerate(ret[metric_group][tgt][metric])
                ]

    if "tune_checkpoint_num" not in ret:
        ret["tune_checkpoint_num"] = 0

    # Upgrades related to extending progress tracker with explicit bests.
    if "checkpoint_number" not in ret:
        ret["checkpoint_number"] = 0

    if "best_eval_metric_steps" not in ret:
        ret["best_eval_metric_steps"] = 0

    if "best_eval_metric_epoch" not in ret:
        ret["best_eval_metric_epoch"] = 0

    if "best_eval_metric_checkpoint_number" not in ret:
        ret["best_eval_metric_checkpoint_number"] = 0

    if "best_eval_train_metrics" not in ret:
        ret["best_eval_train_metrics"] = {}

    if "best_eval_validation_metrics" not in ret:
        ret["best_eval_validation_metrics"] = {}

    if "best_eval_test_metrics" not in ret:
        ret["best_eval_test_metrics"] = {}

    if "best_eval_metric" in ret:
        ret["best_eval_metric_value"] = ret["best_eval_metric"]
        del ret["best_eval_metric"]

    if "last_improvement" in ret:
        del ret["last_improvement"]

    # Delete learning-rate related fields removed in https://github.com/ludwig-ai/ludwig/pull/2877.
    if "best_reduce_learning_rate_eval_metric" in ret:
        del ret["best_reduce_learning_rate_eval_metric"]

    if "last_reduce_learning_rate_eval_metric_improvement" in ret:
        del ret["last_reduce_learning_rate_eval_metric_improvement"]

    return ret


def _traverse_dicts(config: Any, f: Callable[[Dict], None]):
    """Recursively applies function f to every dictionary contained in config.

    f should in-place modify the config dict. f will be called on leaves first, root last.
    """
    if isinstance(config, dict):
        for k, v in config.items():
            _traverse_dicts(v, f)
        f(config)
    elif isinstance(config, list):
        for v in config:
            _traverse_dicts(v, f)


@register_config_transformation("0.6", "backend")
def _update_backend_cache_credentials(backend: Dict[str, Any]) -> Dict[str, Any]:
    if "cache_credentials" in backend:
        credentials = backend.get("credentials", {})
        if "cache" in credentials:
            warnings.warn("`cache` already found in `backend.credentials`, ignoring `cache_credentials`")
        else:
            warnings.warn(
                "`backend.cache_credentials` has been renamed `backend.credentials.cache`", DeprecationWarning
            )
            credentials["cache"] = backend.pop("cache_credentials")
        backend["credentials"] = credentials
    return backend


@register_config_transformation("0.6", ["output_features"])
def update_class_weights_in_features(feature: FeatureConfigDict) -> FeatureConfigDict:
    if LOSS in feature:
        class_weights = feature[LOSS].get(CLASS_WEIGHTS, None)
        if not isinstance(class_weights, list):
            class_weights = None
        feature[LOSS][CLASS_WEIGHTS] = class_weights

    return feature


@register_config_transformation("0.4")
def _update_level_metadata(config: ModelConfigDict) -> ModelConfigDict:
    # Replace parameters represented as keys with params represented as values.
    # Precedence is defined by first in the dictionary order, so if multiple
    # provided keys map to the same value, the one that appears earlier in this
    # dictionary will take priority.
    drop_params = {
        "sequence_length_limit": "max_sequence_length",
        "word_most_common": "most_common",
        "word_sequence_length_limit": "max_sequence_length",
        "word_tokenizer": "tokenizer",
        "word_vocab_file": "vocab_file",
        "char_most_common": "most_common",
        "char_sequence_length_limit": "max_sequence_length",
        "char_tokenizer": "tokenizer",
        "char_vocab_file": "vocab_file",
    }

    def upgrade_params(params):
        for key, value in drop_params.items():
            if key in params:
                if value in params:
                    warnings.warn(
                        f"Removing deprecated config preprocessing parameter {key} as new param {value} already "
                        f"present in the config",
                        DeprecationWarning,
                    )
                else:
                    warnings.warn(
                        f"Renaming deprecated config preprocessing parameter {key} to {value}",
                        DeprecationWarning,
                    )
                    params[value] = params[key]
                del params[key]

    sequence_types = [SEQUENCE, TEXT, AUDIO, TIMESERIES]
    for dtype in sequence_types:
        params = config.get(PREPROCESSING, {}).get(dtype, {})
        upgrade_params(params)

    for feature in config[INPUT_FEATURES]:
        if feature.get(TYPE) not in sequence_types:
            continue
        params = feature.get(PREPROCESSING, {})
        upgrade_params(params)

    return config


@register_config_transformation("0.5")
def rename_training_to_trainer(config: ModelConfigDict) -> ModelConfigDict:
    if TRAINING in config:
        warnings.warn('Config section "training" renamed to "trainer" and will be removed in v0.6', DeprecationWarning)
        config[TRAINER] = config[TRAINING]
        del config[TRAINING]
    return config


@register_config_transformation("0.5", ["input_features", "output_features"])
def _upgrade_use_bias_in_features(feature: FeatureConfigDict) -> FeatureConfigDict:
    def upgrade_use_bias(config):
        if BIAS in config:
            warnings.warn('Parameter "bias" renamed to "use_bias" and will be removed in v0.6', DeprecationWarning)
            config[USE_BIAS] = config[BIAS]
            del config[BIAS]
        if CONV_BIAS in config:
            warnings.warn(
                'Parameter "conv_bias" renamed to "conv_use_bias" and will be removed in v0.6', DeprecationWarning
            )
            config[CONV_USE_BIAS] = config[CONV_BIAS]
            del config[CONV_BIAS]
        if DEFAULT_BIAS in config:
            warnings.warn(
                'Parameter "default_bias" renamed to "default_use_bias" and will be removed in v0.6', DeprecationWarning
            )
            config[DEFAULT_USE_BIAS] = config[DEFAULT_BIAS]
            del config[DEFAULT_BIAS]

    _traverse_dicts(feature, upgrade_use_bias)
    return feature


@register_config_transformation("0.5", ["input_features", "output_features"])
def _upgrade_feature(feature: FeatureConfigDict) -> FeatureConfigDict:
    """Upgrades feature config (in-place)"""
    if feature.get(TYPE) == "numerical":
        warnings.warn('Feature type "numerical" renamed to "number" and will be removed in v0.6', DeprecationWarning)
        feature[TYPE] = NUMBER
    if feature.get(TYPE) == AUDIO:
        if PREPROCESSING in feature:
            feature[PREPROCESSING] = upgrade_audio_preprocessing(feature[PREPROCESSING])
        warnings.warn(
            "Parameters specified at the `audio_feature` parameter level have been unnested and should now "
            "be specified at the preprocessing level. Support for `audio_feature` will be removed in v0.7",
            DeprecationWarning,
        )
    return feature


def upgrade_audio_preprocessing(preproc_dict: PreprocessingConfigDict) -> PreprocessingConfigDict:
    if "audio_feature" in preproc_dict:
        for k, v in preproc_dict["audio_feature"].items():
            preproc_dict[k] = v
        del preproc_dict["audio_feature"]
    return preproc_dict


@register_config_transformation("0.6", ["input_features"])
def _upgrade_encoder_params(feature: FeatureConfigDict) -> FeatureConfigDict:
    return _upgrade_encoder_decoder_params(feature, True)


@register_config_transformation("0.6", ["output_features"])
def _upgrade_decoder_params(feature: FeatureConfigDict) -> FeatureConfigDict:
    return _upgrade_encoder_decoder_params(feature, False)


def _upgrade_encoder_decoder_params(feature: FeatureConfigDict, input_feature: bool) -> FeatureConfigDict:
    """
    This function nests un-nested encoder/decoder parameters to conform with the new config structure for 0.6
    Args:
        feature (Dict): Feature to nest encoder/decoder params for.
        input_feature (Bool): Whether this feature is an input feature or not.
    """
    if TYPE not in feature:
        return feature

    try:
        if input_feature:
            module_type = ENCODER
            feature_cls = get_from_registry(feature[TYPE], get_input_type_registry())
        else:
            module_type = DECODER
            feature_cls = get_from_registry(feature[TYPE], get_output_type_registry())
    except ValueError:
        logging.exception("Failed to obtain encoder / decoder from registry")
        return feature

    feature_schema_cls = feature_cls.get_schema_cls()
    feature_keys = feature_schema_cls.get_valid_field_names()

    # These keys have been renamed from the form below to `fc_<key>` in the new config
    fc_layer_keys = [
        "fc_layers",
        "output_size",
        "use_bias",
        "weights_initializer",
        "bias_initializer",
        "norm",
        "norm_params",
        "activation",
        "dropout",
    ]

    module = feature.get(module_type, {})

    warn = False
    if isinstance(module, str):
        module = {TYPE: module}
        feature[module_type] = module
        warn = True

    nested_params = []
    for k, v in feature.items():
        if k not in feature_keys:
            module[k] = v
            if k in fc_layer_keys and module_type == DECODER:
                module[f"fc_{k}"] = v
            nested_params.append(k)
            warn = True

    if module:
        if module_type in feature:
            feature[module_type].update(module)
        else:
            feature[module_type] = module

    for k in nested_params:
        del feature[k]

    if warn:
        warnings.warn(
            f"{module_type} specific parameters should now be nested within a dictionary under the '{module_type}' "
            f"parameter. Support for un-nested {module_type} specific parameters will be removed in v0.7",
            DeprecationWarning,
        )
    return feature


@register_config_transformation("0.5", ["hyperopt"])
def _upgrade_hyperopt(hyperopt: HyperoptConfigDict) -> HyperoptConfigDict:
    """Upgrades hyperopt config (in-place)"""
    # check for use of legacy "training" reference, if any found convert to "trainer"
    if PARAMETERS in hyperopt:
        hparams = hyperopt[PARAMETERS]
        for k, v in list(hparams.items()):
            substr = "training."
            if k.startswith(substr):
                warnings.warn(
                    'Config section "training" renamed to "trainer" and will be removed in v0.6', DeprecationWarning
                )
                hparams["trainer." + k[len(substr) :]] = v
                del hparams[k]

    # check for legacy parameters in "executor"
    if EXECUTOR in hyperopt:
        hpexecutor = hyperopt[EXECUTOR]
        executor_type = hpexecutor.get(TYPE, None)
        if executor_type is not None and executor_type != RAY:
            warnings.warn(
                f'executor type "{executor_type}" not supported, converted to "ray" will be flagged as error '
                "in v0.6",
                DeprecationWarning,
            )
            hpexecutor[TYPE] = RAY

        # if search_alg not at top level and is present in executor, promote to top level
        if SEARCH_ALG in hpexecutor:
            # promote only if not in top-level, otherwise use current top-level
            if SEARCH_ALG not in hyperopt:
                hyperopt[SEARCH_ALG] = hpexecutor[SEARCH_ALG]
                if isinstance(hyperopt[SEARCH_ALG], str):
                    hyperopt[SEARCH_ALG] = {TYPE: hyperopt[SEARCH_ALG]}
            del hpexecutor[SEARCH_ALG]
    else:
        warnings.warn(
            'Missing "executor" section, adding "ray" executor will be flagged as error in v0.6', DeprecationWarning
        )
        hyperopt[EXECUTOR] = {TYPE: RAY}

    # check for legacy "sampler" section
    if SAMPLER in hyperopt:
        warnings.warn(
            f'"{SAMPLER}" is no longer supported, converted to "{SEARCH_ALG}". "{SAMPLER}" will be flagged as '
            "error in v0.6",
            DeprecationWarning,
        )
        if SEARCH_ALG in hyperopt[SAMPLER]:
            if SEARCH_ALG not in hyperopt:
                hyperopt[SEARCH_ALG] = hyperopt[SAMPLER][SEARCH_ALG]
                if isinstance(hyperopt[SEARCH_ALG], str):
                    hyperopt[SEARCH_ALG] = {TYPE: hyperopt[SEARCH_ALG]}
                warnings.warn('Moved "search_alg" to hyperopt config top-level', DeprecationWarning)

        # if num_samples or scheduler exist in SAMPLER move to EXECUTOR Section
        if NUM_SAMPLES in hyperopt[SAMPLER] and NUM_SAMPLES not in hyperopt[EXECUTOR]:
            hyperopt[EXECUTOR][NUM_SAMPLES] = hyperopt[SAMPLER][NUM_SAMPLES]
            warnings.warn('Moved "num_samples" from "sampler" to "executor"', DeprecationWarning)

        if SCHEDULER in hyperopt[SAMPLER] and SCHEDULER not in hyperopt[EXECUTOR]:
            hyperopt[EXECUTOR][SCHEDULER] = hyperopt[SAMPLER][SCHEDULER]
            warnings.warn('Moved "scheduler" from "sampler" to "executor"', DeprecationWarning)

        if SCHEDULER in hyperopt[EXECUTOR] and len(hyperopt[EXECUTOR][SCHEDULER].keys()) == 0:
            del hyperopt[EXECUTOR][SCHEDULER]

        # remove legacy section
        del hyperopt[SAMPLER]

    if SEARCH_ALG not in hyperopt:
        # make top-level as search_alg, if missing put in default value
        hyperopt[SEARCH_ALG] = {TYPE: "variant_generator"}
        warnings.warn(
            'Missing "search_alg" at hyperopt top-level, adding in default value, will be flagged as error ' "in v0.6",
            DeprecationWarning,
        )
    return hyperopt


@register_config_transformation("0.5", ["trainer"])
def _upgrade_trainer(trainer: TrainerConfigDict) -> TrainerConfigDict:
    """Upgrades trainer config (in-place)"""
    eval_batch_size = trainer.get(EVAL_BATCH_SIZE)
    if eval_batch_size == 0:
        warnings.warn(
            "`trainer.eval_batch_size` value `0` changed to `None`, will be unsupported in v0.6", DeprecationWarning
        )
        trainer[EVAL_BATCH_SIZE] = None
    return trainer


@register_config_transformation("0.5")
def _upgrade_preprocessing_defaults(config: ModelConfigDict) -> ModelConfigDict:
    """Move feature-specific preprocessing parameters into defaults in config (in-place)"""
    type_specific_preprocessing_params = dict()

    # If preprocessing section specified and it contains feature specific preprocessing parameters,
    # make a copy and delete it from the preprocessing section
    for parameter in list(config.get(PREPROCESSING, {})):
        if parameter in get_base_type_registry():
            warnings.warn(
                f"Moving preprocessing configuration for `{parameter}` feature type from `preprocessing` section"
                " to `defaults` section in Ludwig config. This will be unsupported in v0.8.",
                DeprecationWarning,
            )
            type_specific_preprocessing_params[parameter] = config[PREPROCESSING].pop(parameter)

        if parameter == "numerical":
            warnings.warn(
                f"Moving preprocessing configuration for `{parameter}` feature type from `preprocessing` section"
                " to `defaults` section in Ludwig config. This will be unsupported in v0.8.",
                DeprecationWarning,
            )
            type_specific_preprocessing_params[NUMBER] = config[PREPROCESSING].pop(parameter)

    # Delete empty preprocessing section if no other preprocessing parameters specified
    if PREPROCESSING in config and not config[PREPROCESSING]:
        del config[PREPROCESSING]

    # Update defaults with the default feature specific preprocessing parameters
    defaults = config.get(DEFAULTS, {})
    for feature_type, preprocessing_param in type_specific_preprocessing_params.items():
        if PREPROCESSING in preprocessing_param:
            preprocessing_param = preprocessing_param[PREPROCESSING]

        if feature_type == AUDIO:
            preprocessing_param = upgrade_audio_preprocessing(preprocessing_param)

        # If defaults was empty, then create a new key with feature type
        if feature_type not in defaults:
            defaults[feature_type] = {PREPROCESSING: preprocessing_param}
        # Feature type exists but preprocessing hasn't be specified
        elif PREPROCESSING not in defaults[feature_type]:
            defaults[feature_type][PREPROCESSING] = preprocessing_param
        # Update default feature specific preprocessing with parameters from config
        else:
            defaults[feature_type][PREPROCESSING].update(
                merge_dict(defaults[feature_type][PREPROCESSING], preprocessing_param)
            )

    if defaults:
        config[DEFAULTS] = defaults

    return config


@register_config_transformation("0.5", "preprocessing")
def _upgrade_preprocessing_split(preprocessing: PreprocessingConfigDict) -> PreprocessingConfigDict:
    """Upgrade split related parameters in preprocessing."""
    split_params = {}

    force_split = preprocessing.pop(FORCE_SPLIT, None)
    split_probabilities = preprocessing.pop(SPLIT_PROBABILITIES, None)
    stratify = preprocessing.pop(STRATIFY, None)

    if split_probabilities is not None:
        split_params[PROBABILITIES] = split_probabilities
        warnings.warn(
            "`preprocessing.split_probabilities` has been replaced by `preprocessing.split.probabilities`, "
            "will be flagged as error in v0.7",
            DeprecationWarning,
        )

    if stratify is not None:
        split_params[TYPE] = STRATIFY
        split_params[COLUMN] = stratify
        warnings.warn(
            "`preprocessing.stratify` has been replaced by `preprocessing.split.column` "
            'when setting `preprocessing.split.type` to "stratify", '
            "will be flagged as error in v0.7",
            DeprecationWarning,
        )

    if force_split is not None:
        warnings.warn(
            "`preprocessing.force_split` has been replaced by `preprocessing.split.type`, "
            "will be flagged as error in v0.7",
            DeprecationWarning,
        )

        if TYPE not in split_params:
            split_params[TYPE] = RANDOM

    if split_params:
        preprocessing[SPLIT] = split_params

    if AUDIO in preprocessing:
        if "audio_feature" in preprocessing[AUDIO]:
            for k, v in preprocessing[AUDIO]["audio_feature"].items():
                preprocessing[AUDIO][k] = v
            del preprocessing[AUDIO]["audio_feature"]
        warnings.warn(
            "Parameters specified at the `audio_feature` parameter level have been unnested and should now "
            "be specified at the preprocessing level. Support for `audio_feature` will be removed in v0.7",
            DeprecationWarning,
        )
    return preprocessing


@register_config_transformation("0.5")
def update_training(config: ModelConfigDict) -> ModelConfigDict:
    if TRAINING in config:
        warnings.warn('Config section "training" renamed to "trainer" and will be removed in v0.6', DeprecationWarning)
        config[TRAINER] = config[TRAINING]
        del config[TRAINING]
    return config


@register_config_transformation("0.6")
def upgrade_missing_value_strategy(config: ModelConfigDict) -> ModelConfigDict:
    for input_feature in config.get(INPUT_FEATURES, []):
        if _is_old_missing_value_strategy(input_feature):
            _update_old_missing_value_strategy(input_feature)

    for output_feature in config.get(OUTPUT_FEATURES, []):
        if _is_old_missing_value_strategy(output_feature):
            _update_old_missing_value_strategy(output_feature)

    for feature, feature_defaults in config.get(DEFAULTS, {}).items():
        if _is_old_missing_value_strategy(feature_defaults):
            _update_old_missing_value_strategy(config.get(DEFAULTS).get(feature))

    return config


@register_config_transformation("0.6", ["trainer"])
def _upgrade_max_batch_size(trainer: TrainerConfigDict) -> TrainerConfigDict:
    if "increase_batch_size_on_plateau_max" in trainer:
        warnings.warn(
            'Config param "increase_batch_size_on_plateau_max" renamed to "max_batch_size" and will be '
            "removed in v0.8",
            DeprecationWarning,
        )
        increase_batch_size_on_plateau_max_val = trainer.pop("increase_batch_size_on_plateau_max")
        if "max_batch_size" in trainer:
            warnings.warn('"max_batch_size" config param already set. Discarding "increase_batch_size_on_plateau_max".')
        else:
            warnings.warn(
                f'Setting "max_batch_size" config param to "increase_batch_size_on_plateau_max" value '
                f'({increase_batch_size_on_plateau_max_val}) and discarding "increase_batch_size_on_plateau_max"'
            )
            trainer["max_batch_size"] = increase_batch_size_on_plateau_max_val
    return trainer


@register_config_transformation("0.6", ["trainer"])
def remove_trainer_type(trainer: TrainerConfigDict) -> TrainerConfigDict:
    if TYPE in trainer:
        warnings.warn(
            "Config param `type` has been removed from the trainer. The trainer type is determined by the top level "
            " `model_type` parameter. Support for the `type` params in trainer will be removed in v0.8",
            DeprecationWarning,
        )
        del trainer[TYPE]

    return trainer


@register_config_transformation("0.7", ["trainer"])
def learning_rate_scheduler(trainer: TrainerConfigDict) -> TrainerConfigDict:
    key_mapping = {
        "reduce_learning_rate_on_plateau": "reduce_on_plateau",
        "reduce_learning_rate_on_plateau_patience": "reduce_on_plateau_patience",
        "reduce_learning_rate_on_plateau_rate": "reduce_on_plateau_rate",
        "reduce_learning_rate_eval_metric": "reduce_eval_metric",
        "reduce_learning_rate_eval_split": "reduce_eval_split",
        "decay": "decay",
        "decay_steps": "decay_steps",
        "decay_rate": "decay_rate",
        "staircase": "staircase",
        "learning_rate_warmup_epochs": "warmup_evaluations",
    }

    lr_scheduler = trainer.get("learning_rate_scheduler", {})
    for old_key, new_key in key_mapping.items():
        if old_key in trainer:
            warnings.warn(
                f"Config param `trainer.{old_key}` has been moved to `trainer.learning_rate_scheduler.{new_key}`.",
                DeprecationWarning,
            )
            if new_key in lr_scheduler:
                warnings.warn(
                    f"`trainer.learning_rate_scheduler.{new_key}` config param already set. "
                    f"Discarding `trainer.{old_key}`."
                )
            else:
                value = trainer[old_key]
                if old_key == "decay" and isinstance(value, bool):
                    # Decay has changed from a bool to an optional enum
                    lr_scheduler[new_key] = "exponential" if value else None
                elif old_key == "reduce_learning_rate_on_plateau":
                    lr_scheduler[new_key] = int(value)
                else:
                    lr_scheduler[new_key] = value
            del trainer[old_key]

    if lr_scheduler:
        trainer["learning_rate_scheduler"] = lr_scheduler

    return trainer


@register_config_transformation("0.7", ["input_features"])
def _upgrade_legacy_image_encoders(feature: FeatureConfigDict) -> FeatureConfigDict:
    if feature.get(TYPE) != IMAGE:
        return feature

    encoder_mapping = {
        "resnet": "_resnet_legacy",
        "vit": "_vit_legacy",
    }

    encoder = feature.get(ENCODER, {})
    encoder_type = encoder.get(TYPE)
    if encoder_type not in encoder_mapping:
        return feature

    # For this version of Ludwig, only ECD supported these encoders.
    new_encoder_cls = get_encoder_cls(MODEL_ECD, feature[TYPE], encoder_type)
    new_encoder_fields = new_encoder_cls.get_valid_field_names()

    legacy_encoder_cls = get_encoder_cls(MODEL_ECD, feature[TYPE], encoder_mapping[encoder_type])
    legacy_encoder_fields = legacy_encoder_cls.get_valid_field_names()

    user_fields = set(encoder.keys())
    user_fields.remove(TYPE)

    removed_fields = legacy_encoder_fields.difference(new_encoder_fields)
    added_fields = new_encoder_fields.difference(legacy_encoder_fields)

    user_legacy_fields = user_fields.intersection(removed_fields)
    user_new_fields = user_fields.intersection(added_fields)
    if len(user_legacy_fields) > 0:
        if len(user_new_fields) > 0:
            raise ValueError(
                f"Intended encoder type is ambiguous. "
                f"Provided encoder fields matching encoder '{encoder_type}' {user_new_fields} and "
                f"legacy encoder '{encoder_mapping[encoder_type]}' {user_legacy_fields}. "
                f"Please remove features unique to one of these encoder types from your configuration."
            )

        warnings.warn(
            f"Encoder '{encoder_type}' with params '{user_legacy_fields}' has been renamed to "
            f"'{encoder_mapping[encoder_type]}'. Please upgrade your config to use the new '{encoder_type}' as "
            f"support for '{encoder_mapping[encoder_type]}' is not guaranteed past v0.8.",
            DeprecationWarning,
        )

        # User provided legacy fields and no new fields, so we assume they intended to use the legacy encoder
        encoder[TYPE] = encoder_mapping[encoder_type]

    return feature


@register_config_transformation("0.7")
def upgrade_missing_hyperopt(config: ModelConfigDict) -> ModelConfigDict:
    hyperopt = config.get(HYPEROPT)
    if hyperopt == {}:
        # This is a deprecated form of providing a missing hyperopt section, as it violates the schema definition
        warnings.warn(
            "Config section `hyperopt: {}` is deprecated, please set `hyperopt: null` to disable hyperopt.",
            DeprecationWarning,
        )
        del config[HYPEROPT]
    return config


@register_config_transformation("0.7")
def upgrade_defaults_config_for_gbm(config: ModelConfigDict) -> ModelConfigDict:
    model_type = config.get(MODEL_TYPE, "")
    if model_type != MODEL_GBM:
        return config

    defaults_ref = config.get(DEFAULTS, {})
    defaults = copy.deepcopy(defaults_ref)
    gbm_feature_types = GBMDefaultsConfig.Schema().fields.keys()
    for feature_type in defaults_ref:
        if feature_type not in gbm_feature_types:
            del defaults[feature_type]
    config[DEFAULTS] = defaults
    return config


def upgrade_metadata(metadata: TrainingSetMetadataDict) -> TrainingSetMetadataDict:
    # TODO(travis): stopgap solution, we should make it so we don't need to do this
    # by decoupling config and metadata
    metadata = copy.deepcopy(metadata)
    _upgrade_metadata_missing_values(metadata)
    return metadata


def _upgrade_metadata_missing_values(metadata: TrainingSetMetadataDict):
    for k, v in metadata.items():
        if isinstance(v, dict) and _is_old_missing_value_strategy(v):
            _update_old_missing_value_strategy(v)
        elif isinstance(v, dict) and _is_image_feature(v):
            _update_old_image_preprocessing(v)


def _update_old_missing_value_strategy(feature_config: FeatureConfigDict):
    missing_value_strategy = feature_config.get(PREPROCESSING).get(MISSING_VALUE_STRATEGY)
    replacement_strategy = "bfill" if missing_value_strategy == "backfill" else "ffill"
    feature_name = feature_config.get(NAME)
    warnings.warn(
        f"Using `{replacement_strategy}` instead of `{missing_value_strategy}` as the missing value strategy"
        f" for `{feature_name}`. These are identical. `{missing_value_strategy}` will be removed in v0.8",
        DeprecationWarning,
    )
    feature_config[PREPROCESSING].update({MISSING_VALUE_STRATEGY: replacement_strategy})


def _is_old_missing_value_strategy(feature_config: FeatureConfigDict):
    if PREPROCESSING not in feature_config:
        return False
    missing_value_strategy = feature_config.get(PREPROCESSING).get(MISSING_VALUE_STRATEGY, None)
    if not missing_value_strategy or missing_value_strategy not in ("backfill", "pad"):
        return False
    return True


def _is_image_feature(feature_config: FeatureConfigDict):
    preproc = feature_config.get(PREPROCESSING, {})
    return HEIGHT in preproc and WIDTH in preproc


def _update_old_image_preprocessing(feature_config: FeatureConfigDict):
    preprocessing = feature_config.get(PREPROCESSING)
    if not preprocessing:
        return
    preprocessing["standardize_image"] = preprocessing.get("standardize_image")



from typing import Any, Dict

from ludwig.api_annotations import DeveloperAPI
from ludwig.benchmarking.profiler import LudwigProfiler
from ludwig.callbacks import Callback
from ludwig.constants import EVALUATION, PREPROCESSING, TRAINING


# TODO: Change annotation to PublicAPI once Ludwig 0.7 is released
@DeveloperAPI
class LudwigProfilerCallback(Callback):
    """Class that defines the methods necessary to hook into process."""

    def __init__(self, experiment: Dict[str, Any]):
        self.experiment_name = experiment["experiment_name"]
        self.use_torch_profiler = experiment["profiler"]["use_torch_profiler"]
        self.logging_interval = experiment["profiler"]["logging_interval"]
        self.preprocess_profiler = None
        self.train_profiler = None
        self.evaluation_profiler = None

    def on_preprocess_start(self, *args, **kwargs):
        self.preprocess_profiler = LudwigProfiler(
            tag=PREPROCESSING,
            output_dir=self.experiment_name,
            use_torch_profiler=self.use_torch_profiler,
            logging_interval=self.logging_interval,
        )
        self.preprocess_profiler.__enter__()

    def on_preprocess_end(self, *args, **kwargs):
        self.preprocess_profiler.__exit__(None, None, None)
        del self.preprocess_profiler

    def on_train_start(self, *args, **kwargs):
        self.train_profiler = LudwigProfiler(
            tag=TRAINING,
            output_dir=self.experiment_name,
            use_torch_profiler=self.use_torch_profiler,
            logging_interval=self.logging_interval,
        )
        self.train_profiler.__enter__()

    def on_train_end(self, *args, **kwargs):
        self.train_profiler.__exit__(None, None, None)
        del self.train_profiler

    def on_evaluation_start(self):
        self.evaluation_profiler = LudwigProfiler(
            tag=EVALUATION,
            output_dir=self.experiment_name,
            use_torch_profiler=self.use_torch_profiler,
            logging_interval=self.logging_interval,
        )
        self.evaluation_profiler.__enter__()

    def on_evaluation_end(self):
        self.evaluation_profiler.__exit__(None, None, None)
        del self.evaluation_profiler



_logged = set()


def log_once(key: str) -> bool:
    """Returns True if this is the "first" call for a given key.

    Example:
        if log_once("some_key"):
            logger.info("Some verbose logging statement") # noqa
    """

    if key not in _logged:
        _logged.add(key)
        return True
    return False



# Copyright (c) 2019 Uber Technologies, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
import logging
from functools import partial
from typing import Any, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn

from ludwig.utils.image_utils import get_img_output_shape
from ludwig.utils.torch_utils import get_activation, LudwigModule

logger = logging.getLogger(__name__)


class Conv1DLayer(LudwigModule):
    def __init__(
        self,
        in_channels=1,
        out_channels=256,
        max_sequence_length=None,
        kernel_size=3,
        strides=1,
        padding="same",
        dilation=1,
        groups=1,
        use_bias=True,
        weights_initializer="xavier_uniform",
        bias_initializer="zeros",
        norm=None,
        norm_params=None,
        activation="relu",
        dropout=0,
        pool_function="max",
        pool_size=2,
        pool_strides=None,
        pool_padding="valid",
    ):
        super().__init__()

        self.in_channels = in_channels
        self.out_channels = out_channels
        self.max_sequence_length = max_sequence_length
        self.kernel_size = kernel_size
        self.stride = strides
        self.padding = padding
        self.dilation = dilation
        self.groups = groups
        self.pool_size = pool_size
        if pool_strides is None:
            self.pool_strides = pool_size
        else:
            self.pool_strides = pool_strides
        if pool_padding == "same" and pool_size is not None:
            self.pool_padding = (self.pool_size - 1) // 2
        else:
            self.pool_padding = 0

        self.layers = nn.ModuleList()

        self.layers.append(
            nn.Conv1d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=(kernel_size,),
                stride=(strides,),
                padding=padding,
                dilation=(dilation,),
            )
        )

        if norm and norm_params is None:
            norm_params = {}
        if norm == "batch":
            self.layers.append(nn.BatchNorm1d(num_features=out_channels, **norm_params))
        elif norm == "layer":
            self.layers.append(nn.LayerNorm(normalized_shape=[out_channels, self.max_sequence_length], **norm_params))

        self.layers.append(get_activation(activation))

        if dropout > 0:
            self.layers.append(nn.Dropout(dropout))

        if pool_size is not None:
            pool = nn.MaxPool1d
            if pool_function in {"average", "avg", "mean"}:
                pool = nn.AvgPool1d
            self.layers.append(pool(kernel_size=self.pool_size, stride=self.pool_strides, padding=self.pool_padding))

        for layer in self.layers:
            logger.debug(f"   {layer._get_name()}")

    @property
    def input_shape(self):
        """Returns the size of the input tensor without the batch dimension."""
        return torch.Size([self.max_sequence_length, self.in_channels])

    def forward(self, inputs, training=None, mask=None):
        # inputs: [batch_size, seq_size, in_channels]
        # in Torch nomenclature (N, L, C)
        hidden = inputs

        # put in torch compatible form [batch_size, in_channels, seq_size]
        hidden = hidden.transpose(1, 2)

        for layer in self.layers:
            hidden = layer(hidden)

        # revert back to normal form [batch_size, seq_size, out_channels]
        hidden = hidden.transpose(1, 2)

        return hidden  # (batch_size, seq_size, out_channels)


class Conv1DStack(LudwigModule):
    def __init__(
        self,
        in_channels=1,
        max_sequence_length=None,
        layers=None,
        num_layers=None,
        default_num_filters=256,
        default_filter_size=3,
        default_strides=1,
        default_padding="same",
        default_dilation_rate=1,
        default_use_bias=True,
        default_weights_initializer="xavier_uniform",
        default_bias_initializer="zeros",
        default_norm=None,
        default_norm_params=None,
        default_activation="relu",
        default_dropout=0,
        default_pool_function="max",
        default_pool_size=2,
        default_pool_strides=None,
        default_pool_padding="same",
        **kwargs,
    ):
        super().__init__()

        self.max_sequence_length = max_sequence_length
        self.in_channels = in_channels

        if layers is None:
            if num_layers is None:
                self.layers = [
                    {"filter_size": 7, "pool_size": 3},
                    {"filter_size": 7, "pool_size": 3},
                    {"filter_size": 3, "pool_size": None},
                    {"filter_size": 3, "pool_size": None},
                    {"filter_size": 3, "pool_size": None},
                    {"filter_size": 3, "pool_size": 3},
                ]
            else:
                self.layers = []
                for i in range(num_layers):
                    self.layers.append(
                        {
                            "filter_size": default_filter_size,
                            "num_filters": default_num_filters,
                            "pool_size": default_pool_size,
                            "pool_strides": default_pool_strides,
                        }
                    )
        else:
            self.layers = layers

        for layer in self.layers:
            if "num_filters" not in layer:
                layer["num_filters"] = default_num_filters
            if "filter_size" not in layer:
                layer["filter_size"] = default_filter_size
            if "strides" not in layer:
                layer["strides"] = default_strides
            if "padding" not in layer:
                layer["padding"] = default_padding
            if "dilation_rate" not in layer:
                layer["dilation_rate"] = default_dilation_rate
            if "use_bias" not in layer:
                layer["use_bias"] = default_use_bias
            if "weights_initializer" not in layer:
                layer["weights_initializer"] = default_weights_initializer
            if "bias_initializer" not in layer:
                layer["bias_initializer"] = default_bias_initializer
            if "norm" not in layer:
                layer["norm"] = default_norm
            if "norm_params" not in layer:
                layer["norm_params"] = default_norm_params
            if "activation" not in layer:
                layer["activation"] = default_activation
            if "dropout" not in layer:
                layer["dropout"] = default_dropout
            if "pool_function" not in layer:
                layer["pool_function"] = default_pool_function
            if "pool_size" not in layer:
                layer["pool_size"] = default_pool_size
            if "pool_strides" not in layer:
                layer["pool_strides"] = default_pool_strides
            if "pool_padding" not in layer:
                layer["pool_padding"] = default_pool_padding

        self.stack = nn.ModuleList()

        prior_layer_channels = in_channels
        l_in = self.max_sequence_length  # torch L_in
        for i, layer in enumerate(self.layers):
            logger.debug(f"   stack layer {i}")
            self.stack.append(
                Conv1DLayer(
                    in_channels=prior_layer_channels,
                    out_channels=layer["num_filters"],
                    max_sequence_length=l_in,
                    kernel_size=layer["filter_size"],
                    strides=layer["strides"],
                    padding=layer["padding"],
                    dilation=layer["dilation_rate"],
                    use_bias=layer["use_bias"],
                    weights_initializer=layer["weights_initializer"],
                    bias_initializer=layer["bias_initializer"],
                    norm=layer["norm"],
                    norm_params=layer["norm_params"],
                    activation=layer["activation"],
                    dropout=layer["dropout"],
                    pool_function=layer["pool_function"],
                    pool_size=layer["pool_size"],
                    pool_strides=layer["pool_strides"],
                    pool_padding=layer["pool_padding"],
                )
            )

            # retrieve number of channels from prior layer
            input_shape = self.stack[i].input_shape
            output_shape = self.stack[i].output_shape

            logger.debug(f"{self.__class__.__name__}: " f"input_shape {input_shape}, output shape {output_shape}")

            # pass along shape for the input to the next layer
            l_in, prior_layer_channels = output_shape

    @property
    def input_shape(self):
        """Returns the size of the input tensor without the batch dimension."""
        return torch.Size([self.max_sequence_length, self.in_channels])

    def forward(self, inputs, mask=None):
        hidden = inputs

        # todo: enumerate for debugging, remove after testing
        for i, layer in enumerate(self.stack):
            hidden = layer(hidden)

        if hidden.shape[1] == 0:
            raise ValueError(
                "The output of the conv stack has the second dimension "
                "(length of the sequence) equal to 0. "
                "This means that the combination of filter_size, padding, "
                "stride, pool_size, pool_padding and pool_stride reduces "
                "the sequence length more than is possible. "
                'Try using "same" padding and reducing or eliminating stride '
                "and pool."
            )

        return hidden


class ParallelConv1D(LudwigModule):
    def __init__(
        self,
        in_channels=1,
        max_sequence_length=None,
        layers=None,
        default_num_filters=256,
        default_filter_size=3,
        default_strides=1,
        default_padding="same",
        default_dilation_rate=1,
        default_use_bias=True,
        default_weights_initializer="xavier_uniform",
        default_bias_initializer="zeros",
        default_norm=None,
        default_norm_params=None,
        default_activation="relu",
        default_dropout=0,
        default_pool_function="max",
        default_pool_size=None,
        default_pool_strides=None,
        default_pool_padding="valid",
        **kwargs,
    ):
        super().__init__()

        self.in_channels = in_channels
        self.max_sequence_length = max_sequence_length

        if layers is None:
            self.layers = [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}]
        else:
            self.layers = layers

        for layer in self.layers:
            if "num_filters" not in layer:
                layer["num_filters"] = default_num_filters
            if "filter_size" not in layer:
                layer["filter_size"] = default_filter_size
            if "strides" not in layer:
                layer["strides"] = default_strides
            if "padding" not in layer:
                layer["padding"] = default_padding
            if "dilation_rate" not in layer:
                layer["dilation_rate"] = default_dilation_rate
            if "use_bias" not in layer:
                layer["use_bias"] = default_use_bias
            if "weights_initializer" not in layer:
                layer["weights_initializer"] = default_weights_initializer
            if "bias_initializer" not in layer:
                layer["bias_initializer"] = default_bias_initializer
            if "norm" not in layer:
                layer["norm"] = default_norm
            if "norm_params" not in layer:
                layer["norm_params"] = default_norm_params
            if "activation" not in layer:
                layer["activation"] = default_activation
            if "dropout" not in layer:
                layer["dropout"] = default_dropout
            if "pool_function" not in layer:
                layer["pool_function"] = default_pool_function
            if "pool_size" not in layer:
                layer["pool_size"] = default_pool_size
            if "pool_strides" not in layer:
                layer["pool_strides"] = default_pool_strides
            if "pool_padding" not in layer:
                layer["pool_padding"] = default_pool_padding

        self.parallel_layers = nn.ModuleList()

        for i, layer in enumerate(self.layers):
            logger.debug(f"   parallel layer {i}")
            self.parallel_layers.append(
                Conv1DLayer(
                    in_channels=self.in_channels,
                    out_channels=layer["num_filters"],
                    max_sequence_length=self.max_sequence_length,
                    kernel_size=layer["filter_size"],
                    strides=layer["strides"],
                    padding=layer["padding"],
                    dilation=layer["dilation_rate"],
                    use_bias=layer["use_bias"],
                    weights_initializer=layer["weights_initializer"],
                    bias_initializer=layer["bias_initializer"],
                    norm=layer["norm"],
                    norm_params=layer["norm_params"],
                    activation=layer["activation"],
                    dropout=layer["dropout"],
                    pool_function=layer["pool_function"],
                    pool_size=layer["pool_size"],
                    pool_strides=layer["pool_strides"],
                    pool_padding=layer["pool_padding"],
                )
            )

            logger.debug(
                f"{self.__class__.__name__} layer {i}, input shape "
                f"{self.parallel_layers[i].input_shape}, output shape "
                f"{self.parallel_layers[i].output_shape}"
            )

    @property
    def input_shape(self) -> torch.Size:
        """Returns the size of the input tensor without the batch dimension."""
        return torch.Size([self.max_sequence_length, self.in_channels])

    def forward(self, inputs, mask=None):
        # inputs: [batch_size, seq_size, in_channels)

        hidden = inputs
        hiddens = []

        for layer in self.parallel_layers:
            hiddens.append(layer(hidden))
        hidden = torch.cat(hiddens, 2)

        if hidden.shape[1] == 0:
            raise ValueError(
                "The output of the conv stack has the second dimension "
                "(length of the sequence) equal to 0. "
                "This means that the combination of filter_size, padding, "
                "stride, pool_size, pool_padding and pool_stride reduces "
                "the sequence length more than is possible. "
                'Try using "same" padding and reducing or eliminating stride '
                "and pool."
            )

        # (batch_size, seq_size, len(parallel_layers) * out_channels)
        return hidden


class ParallelConv1DStack(LudwigModule):
    def __init__(
        self,
        in_channels=None,
        stacked_layers=None,
        max_sequence_length=None,
        default_num_filters=64,
        default_filter_size=3,
        default_strides=1,
        default_padding="same",
        default_dilation_rate=1,
        default_use_bias=True,
        default_weights_initializer="xavier_uniform",
        default_bias_initializer="zeros",
        default_norm=None,
        default_norm_params=None,
        default_activation="relu",
        default_dropout=0,
        default_pool_function="max",
        default_pool_size=None,
        default_pool_strides=None,
        default_pool_padding="valid",
        **kwargs,
    ):
        super().__init__()

        self.max_sequence_length = max_sequence_length
        self.in_channels = in_channels

        if stacked_layers is None:
            self.stacked_parallel_layers = [
                [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}],
                [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}],
                [{"filter_size": 2}, {"filter_size": 3}, {"filter_size": 4}, {"filter_size": 5}],
            ]

        else:
            self.stacked_parallel_layers = stacked_layers

        for i, parallel_layers in enumerate(self.stacked_parallel_layers):
            for j in range(len(parallel_layers)):
                layer = parallel_layers[j]
                if "num_filters" not in layer:
                    layer["num_filters"] = default_num_filters
                if "filter_size" not in layer:
                    layer["filter_size"] = default_filter_size
                if "strides" not in layer:
                    layer["strides"] = default_strides
                if "padding" not in layer:
                    layer["padding"] = default_padding
                if "dilation_rate" not in layer:
                    layer["dilation_rate"] = default_dilation_rate
                if "use_bias" not in layer:
                    layer["use_bias"] = default_use_bias
                if "weights_initializer" not in layer:
                    layer["weights_initializer"] = default_weights_initializer
                if "bias_initializer" not in layer:
                    layer["bias_initializer"] = default_bias_initializer
                if "norm" not in layer:
                    layer["norm"] = default_norm
                if "norm_params" not in layer:
                    layer["norm_params"] = default_norm_params
                if "activation" not in layer:
                    layer["activation"] = default_activation
                if "dropout" not in layer:
                    layer["dropout"] = default_dropout
                if "pool_function" not in layer:
                    layer["pool_function"] = default_pool_function
                if "pool_size" not in layer:
                    if i == len(self.stacked_parallel_layers) - 1:
                        layer["pool_size"] = default_pool_size
                    else:
                        layer["pool_size"] = None
                if "pool_strides" not in layer:
                    layer["pool_strides"] = default_pool_strides
                if "pool_padding" not in layer:
                    layer["pool_padding"] = default_pool_padding

        self.stack = nn.ModuleList()
        num_channels = self.in_channels
        sequence_length = self.max_sequence_length
        for i, parallel_layers in enumerate(self.stacked_parallel_layers):
            logger.debug(f"   stack layer {i}")
            self.stack.append(ParallelConv1D(num_channels, sequence_length, layers=parallel_layers))

            logger.debug(
                f"{self.__class__.__name__} layer {i}, input shape "
                f"{self.stack[i].input_shape}, output shape "
                f"{self.stack[i].output_shape}"
            )

            # set input specification for the layer
            num_channels = self.stack[i].output_shape[1]
            sequence_length = self.stack[i].output_shape[0]

    @property
    def input_shape(self):
        """Returns the size of the input tensor without the batch dimension."""
        return torch.Size([self.max_sequence_length, self.in_channels])

    def forward(self, inputs, mask=None):
        hidden = inputs

        for layer in self.stack:
            hidden = layer(hidden)

        if hidden.shape[2] == 0:
            raise ValueError(
                "The output of the conv stack has the second dimension "
                "(length of the sequence) equal to 0. "
                "This means that the combination of filter_size, padding, "
                "stride, pool_size, pool_padding and pool_stride is reduces "
                "the sequence length more than is possible. "
                'Try using "same" padding and reducing or eliminating stride '
                "and pool."
            )

        return hidden


class Conv2DLayer(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        in_channels: int,
        out_channels: int = 256,
        kernel_size: Union[int, Tuple[int]] = 3,
        stride: Union[int, Tuple[int]] = 1,
        padding: Union[int, Tuple[int], str] = "valid",
        dilation: Union[int, Tuple[int]] = 1,
        groups: int = 1,
        use_bias: bool = True,
        padding_mode: str = "zeros",
        norm: Optional[str] = None,
        norm_params: Optional[Dict[str, Any]] = None,
        activation: str = "relu",
        dropout: float = 0,
        pool_function: int = "max",
        pool_kernel_size: Union[int, Tuple[int]] = None,
        pool_stride: Optional[int] = None,
        pool_padding: Union[int, Tuple[int]] = 0,
        pool_dilation: Union[int, Tuple[int]] = 1,
    ):
        super().__init__()

        self.layers = torch.nn.ModuleList()

        self._input_shape = (in_channels, img_height, img_width)
        pool_stride = pool_stride or pool_kernel_size

        self.layers.append(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=groups,
                bias=use_bias,
                padding_mode=padding_mode,
            )
        )
        out_height, out_width = get_img_output_shape(img_height, img_width, kernel_size, stride, padding, dilation)

        if norm and norm_params is None:
            norm_params = {}
        if norm == "batch":
            # Batch norm over channels
            self.layers.append(nn.BatchNorm2d(num_features=out_channels, **norm_params))
        elif norm == "layer":
            # Layer norm over image height and width
            self.layers.append(nn.LayerNorm(normalized_shape=(out_height, out_width), **norm_params))

        self.layers.append(get_activation(activation))

        if dropout > 0:
            self.layers.append(nn.Dropout(dropout))

        if pool_kernel_size is not None:
            pool = partial(nn.MaxPool2d, dilation=pool_dilation)
            if pool_function in {"average", "avg", "mean"}:
                pool = nn.AvgPool2d
            self.layers.append(pool(kernel_size=pool_kernel_size, stride=pool_stride, padding=pool_padding))
            out_height, out_width = get_img_output_shape(
                img_height=out_height,
                img_width=out_width,
                kernel_size=pool_kernel_size,
                stride=pool_stride,
                padding=pool_padding,
                dilation=pool_dilation,
            )

        for layer in self.layers:
            logger.debug(f"   {layer._get_name()}")

        self._output_shape = (out_channels, out_height, out_width)

    def forward(self, inputs):
        hidden = inputs

        for layer in self.layers:
            hidden = layer(hidden)

        return hidden

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)


class Conv2DStack(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        layers: Optional[List[Dict]] = None,
        num_layers: Optional[int] = None,
        first_in_channels: Optional[int] = None,
        default_out_channels: int = 256,
        default_kernel_size: Union[int, Tuple[int]] = 3,
        default_stride: Union[int, Tuple[int]] = 1,
        default_padding: Union[int, Tuple[int], str] = "valid",
        default_dilation: Union[int, Tuple[int]] = 1,
        default_groups: int = 1,
        default_use_bias: bool = True,
        default_padding_mode: str = "zeros",
        default_norm: Optional[str] = None,
        default_norm_params: Optional[Dict[str, Any]] = None,
        default_activation: str = "relu",
        default_dropout: int = 0,
        default_pool_function: int = "max",
        default_pool_kernel_size: Union[int, Tuple[int]] = 2,
        default_pool_stride: Union[int, Tuple[int]] = None,
        default_pool_padding: Union[int, Tuple[int]] = 0,
        default_pool_dilation: Union[int, Tuple[int]] = 1,
    ):
        super().__init__()

        # Confirm that all inputs are consistent
        first_in_channels = self._check_in_channels(first_in_channels, layers)
        default_pool_stride = default_pool_stride or default_pool_kernel_size
        if layers is not None and num_layers is not None:
            raise Warning("Both layers and num_layers are not None." "Default to using layers.")
        if (
            first_in_channels is not None
            and layers is not None
            and len(layers) > 0
            and "in_channels" in layers[0]
            and layers[0]["in_channels"] != first_in_channels
        ):
            raise Warning(
                "Input channels is set via layers[0]['in_channels'] and first_in_channels."
                "Default to using first_in_channels."
            )

        self._input_shape = (first_in_channels, img_height, img_width)

        if layers is None:
            if num_layers is None:
                self.layers = [
                    {"out_channels": 32},
                    {"out_channels": 64},
                ]
            else:
                self.layers = []
                for i in range(num_layers):
                    self.layers.append(
                        {
                            "kernel_size": default_kernel_size,
                            "out_channels": default_out_channels,
                            "pool_kernel_size": default_pool_kernel_size,
                        }
                    )
        else:
            self.layers = layers

        for layer in self.layers:
            if "out_channels" not in layer:
                layer["out_channels"] = default_out_channels
            if "kernel_size" not in layer:
                layer["kernel_size"] = default_kernel_size
            if "stride" not in layer:
                layer["stride"] = default_stride
            if "padding" not in layer:
                layer["padding"] = default_padding
            if "dilation" not in layer:
                layer["dilation"] = default_dilation
            if "groups" not in layer:
                layer["groups"] = default_groups
            if "use_bias" not in layer:
                layer["use_bias"] = default_use_bias
            if "padding_mode" not in layer:
                layer["padding_mode"] = default_padding_mode
            if "norm" not in layer:
                layer["norm"] = default_norm
            if "norm_params" not in layer:
                layer["norm_params"] = default_norm_params
            if "activation" not in layer:
                layer["activation"] = default_activation
            if "dropout" not in layer:
                layer["dropout"] = default_dropout
            if "pool_function" not in layer:
                layer["pool_function"] = default_pool_function
            if "pool_kernel_size" not in layer:
                layer["pool_kernel_size"] = default_pool_kernel_size
            if "pool_stride" not in layer:
                layer["pool_stride"] = default_pool_stride
            if "pool_padding" not in layer:
                layer["pool_padding"] = default_pool_padding
            if "pool_dilation" not in layer:
                layer["pool_dilation"] = default_pool_dilation

        self.stack = torch.nn.ModuleList()

        in_channels = first_in_channels
        for i, layer in enumerate(self.layers):
            logger.debug(f"   stack layer {i}")
            self.stack.append(
                Conv2DLayer(
                    img_height=img_height,
                    img_width=img_width,
                    in_channels=in_channels,
                    out_channels=layer["out_channels"],
                    kernel_size=layer["kernel_size"],
                    stride=layer["stride"],
                    padding=layer["padding"],
                    dilation=layer["dilation"],
                    groups=layer["groups"],
                    use_bias=layer["use_bias"],
                    padding_mode=layer["padding_mode"],
                    norm=layer["norm"],
                    norm_params=layer["norm_params"],
                    activation=layer["activation"],
                    dropout=layer["dropout"],
                    pool_function=layer["pool_function"],
                    pool_kernel_size=layer["pool_kernel_size"],
                    pool_stride=layer["pool_stride"],
                    pool_padding=layer["pool_padding"],
                    pool_dilation=layer["pool_dilation"],
                )
            )
            in_channels, img_height, img_width = self.stack[-1].output_shape

        self._output_shape = (in_channels, img_height, img_width)

    def forward(self, inputs):
        hidden = inputs

        for layer in self.stack:
            hidden = layer(hidden)

        return hidden

    def _check_in_channels(self, first_in_channels: Optional[int], layers: Optional[List[Dict]]) -> None:
        """Confirms that in_channels for first layer of the stack exists."""

        if first_in_channels is not None:
            return first_in_channels
        elif layers is not None and len(layers) > 0 and "in_channels" in layers[0]:
            return layers[0]["in_channels"]
        raise ValueError(
            "In_channels for first layer should be specified either via " "`first_in_channels` or `layers` arguments."
        )

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)

    @property
    def input_shape(self) -> torch.Size:
        return torch.size(self._input_shape)


class Conv2DLayerFixedPadding(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        in_channels: int,
        out_channels=256,
        kernel_size=3,
        stride=1,
        dilation=1,
        groups=1,
        use_bias=False,
    ):
        super().__init__()

        self.layers = torch.nn.ModuleList()
        self._input_shape = (in_channels, img_height, img_width)

        padding = "same"
        if stride > 1:
            padding = (kernel_size - 1) // 2

        self.layers.append(
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=stride,
                padding=padding,
                dilation=dilation,
                groups=groups,
                bias=use_bias,
            )
        )
        img_height, img_width = get_img_output_shape(
            img_height=img_height,
            img_width=img_width,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
        )

        for layer in self.layers:
            logger.debug(f"   {layer._get_name()}")

        self._output_shape = (out_channels, img_height, img_width)

    def forward(self, inputs):
        hidden = inputs

        for layer in self.layers:
            hidden = layer(hidden)

        return hidden

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)


class ResNetBlock(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        first_in_channels: int,
        out_channels: int,
        stride: int = 1,
        batch_norm_momentum: float = 0.1,
        batch_norm_epsilon: float = 0.001,
        projection_shortcut: Optional[LudwigModule] = None,
    ):
        """Resnet blocks used for ResNet34 and smaller.

        stride: A single int specifying the stride of the first convolution.
            The last convolution will have stride of 1.
        """
        super().__init__()
        self._input_shape = (first_in_channels, img_height, img_width)

        self.conv1 = Conv2DLayerFixedPadding(
            img_height=img_height,
            img_width=img_width,
            in_channels=first_in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=stride,
        )
        in_channels, img_height, img_width = self.conv1.output_shape
        self.norm1 = nn.BatchNorm2d(num_features=in_channels, eps=batch_norm_epsilon, momentum=batch_norm_momentum)
        self.relu1 = get_activation("relu")

        self.conv2 = Conv2DLayerFixedPadding(
            img_height=img_height,
            img_width=img_width,
            in_channels=out_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=1,
        )
        self.norm2 = nn.BatchNorm2d(num_features=out_channels, eps=batch_norm_epsilon, momentum=batch_norm_momentum)
        self.relu2 = get_activation("relu")

        for layer in [self.conv1, self.norm1, self.relu1, self.conv2, self.norm2, self.relu2]:
            logger.debug(f"   {layer._get_name()}")

        self._output_shape = self.conv2.output_shape

        self.projection_shortcut = projection_shortcut
        if self.projection_shortcut is not None and self.projection_shortcut.output_shape != self._output_shape:
            raise ValueError(
                f"Output shapes of ResnetBlock and projection_shortcut should "
                f"match but are {self._output_shape} and "
                f"{self.projection_shortcut.output_shape} respectively."
            )
        if self.projection_shortcut is None and self._input_shape != self._output_shape:
            self.projection_shortcut = Conv2DLayer(
                img_height=self._input_shape[1],
                img_width=self._input_shape[2],
                in_channels=first_in_channels,
                out_channels=out_channels,
                kernel_size=1,
                stride=stride,
            )

    def forward(self, inputs):
        shortcut = inputs

        if self.projection_shortcut is not None:
            shortcut = self.projection_shortcut(shortcut)

        hidden = self.conv1(inputs)
        hidden = self.norm1(hidden)
        hidden = self.relu1(hidden)
        hidden = self.conv2(hidden)
        hidden = self.norm2(hidden)

        return self.relu2(hidden + shortcut)

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)


# TODO(shreya): Combine with ResNetBlock by adding a flag.
class ResNetBottleneckBlock(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        first_in_channels: int,
        out_channels: int,
        stride: int = 1,
        batch_norm_momentum: float = 0.1,
        batch_norm_epsilon: float = 0.001,
        projection_shortcut: Optional[LudwigModule] = None,
    ):
        """Resnet bottleneck blocks used for ResNet50 and larger.

        stride: A single int specifying the stride of the middle convolution.
            The first and last convolution will have stride of 1.
        """
        super().__init__()

        self._input_shape = (first_in_channels, img_height, img_width)

        self.conv1 = Conv2DLayerFixedPadding(
            img_height=img_height,
            img_width=img_width,
            in_channels=first_in_channels,
            out_channels=out_channels,
            kernel_size=1,
            stride=1,
        )
        in_channels, img_height, img_width = self.conv1.output_shape
        self.norm1 = nn.BatchNorm2d(num_features=in_channels, eps=batch_norm_epsilon, momentum=batch_norm_momentum)
        self.relu1 = get_activation("relu")

        self.conv2 = Conv2DLayerFixedPadding(
            img_height=img_height,
            img_width=img_width,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=3,
            stride=stride,
        )
        in_channels, img_height, img_width = self.conv2.output_shape
        self.norm2 = nn.BatchNorm2d(num_features=in_channels, eps=batch_norm_epsilon, momentum=batch_norm_momentum)
        self.relu2 = get_activation("relu")

        self.conv3 = Conv2DLayerFixedPadding(
            img_height=img_height,
            img_width=img_width,
            in_channels=in_channels,
            out_channels=4 * out_channels,
            kernel_size=1,
            stride=1,
        )
        self.norm3 = nn.BatchNorm2d(num_features=4 * out_channels, eps=batch_norm_epsilon, momentum=batch_norm_momentum)
        self.relu3 = get_activation("relu")

        for layer in [
            self.conv1,
            self.norm1,
            self.relu1,
            self.conv2,
            self.norm2,
            self.relu2,
            self.conv3,
            self.norm3,
            self.relu3,
        ]:
            logger.debug(f"   {layer._get_name()}")

        self._output_shape = self.conv3.output_shape

        self.projection_shortcut = projection_shortcut
        if self.projection_shortcut is not None and self.projection_shortcut.output_shape != self._output_shape:
            raise ValueError(
                f"Output shapes of ResnetBlock and projection_shortcut should "
                f"match but are {self._output_shape} and "
                f"{self.projection_shortcut.output_shape} respectively."
            )
        if self.projection_shortcut is None and self._input_shape != self._output_shape:
            self.projection_shortcut = Conv2DLayer(
                img_height=self._input_shape[1],
                img_width=self._input_shape[2],
                in_channels=first_in_channels,
                out_channels=4 * out_channels,
                kernel_size=1,
                stride=stride,
            )

    def forward(self, inputs):
        shortcut = inputs

        if self.projection_shortcut is not None:
            shortcut = self.projection_shortcut(shortcut)

        hidden = self.conv1(inputs)
        hidden = self.norm1(hidden)
        hidden = self.relu1(hidden)
        hidden = self.conv2(hidden)
        hidden = self.norm2(hidden)
        hidden = self.relu2(hidden)
        hidden = self.conv3(hidden)
        hidden = self.norm3(hidden)

        return self.relu3(hidden + shortcut)

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)


class ResNetBlockLayer(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        first_in_channels: int,
        out_channels: int,
        is_bottleneck: bool,
        block_fn: Union[ResNetBlock, ResNetBottleneckBlock],
        num_blocks: int,
        stride: Union[int, Tuple[int]] = 1,
        batch_norm_momentum: float = 0.1,
        batch_norm_epsilon: float = 0.001,
    ):
        super().__init__()

        self._input_shape = (first_in_channels, img_height, img_width)

        # Bottleneck blocks end with 4x the number of channels as they start with
        projection_out_channels = out_channels * 4 if is_bottleneck else out_channels
        projection_shortcut = Conv2DLayerFixedPadding(
            img_height=img_height,
            img_width=img_width,
            in_channels=first_in_channels,
            out_channels=projection_out_channels,
            kernel_size=1,
            stride=stride,
        )

        self.layers = torch.nn.ModuleList(
            [
                block_fn(
                    img_height,
                    img_width,
                    first_in_channels,
                    out_channels,
                    stride,
                    batch_norm_momentum,
                    batch_norm_epsilon,
                    projection_shortcut,
                )
            ]
        )
        in_channels, img_height, img_width = self.layers[-1].output_shape

        for _ in range(1, num_blocks):
            self.layers.append(
                block_fn(
                    img_height=img_height,
                    img_width=img_width,
                    first_in_channels=in_channels,
                    out_channels=out_channels,
                    stride=1,
                    batch_norm_momentum=batch_norm_momentum,
                    batch_norm_epsilon=batch_norm_epsilon,
                )
            )
            in_channels, img_height, img_width = self.layers[-1].output_shape

        for layer in self.layers:
            logger.debug(f"   {layer._get_name()}")

        self._output_shape = (in_channels, img_height, img_width)

    def forward(self, inputs):
        hidden = inputs
        for layer in self.layers:
            hidden = layer(hidden)
        return hidden

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)


class ResNet(LudwigModule):
    def __init__(
        self,
        img_height: int,
        img_width: int,
        first_in_channels: int,
        out_channels: int,
        resnet_size: int = 34,
        kernel_size: Union[int, Tuple[int]] = 7,
        conv_stride: Union[int, Tuple[int]] = 2,
        first_pool_kernel_size: Union[int, Tuple[int]] = 3,
        first_pool_stride: Union[int, Tuple[int]] = 2,
        block_sizes: List[int] = None,
        block_strides: List[Union[int, Tuple[int]]] = None,
        batch_norm_momentum: float = 0.1,
        batch_norm_epsilon: float = 0.001,
    ):
        """Creates a model obtaining an image representation.

        Implements ResNet v2:
        Identity Mappings in Deep Residual Networks
        https://arxiv.org/pdf/1603.05027.pdf
        by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Jul 2016.

        Args:
          resnet_size: A single integer for the size of the ResNet model.
          is_bottleneck: Use regular blocks or bottleneck blocks.
          out_channels: The number of filters to use for the first block layer
            of the model. This number is then doubled for each subsequent block
            layer.
          kernel_size: The kernel size to use for convolution.
          conv_stride: stride size for the initial convolutional layer
          first_pool_kernel_size: Pool size to be used for the first pooling layer.
            If none, the first pooling layer is skipped.
          first_pool_stride: stride size for the first pooling layer. Not used
            if first_pool_kernel_size is None.
          block_sizes: A list containing n values, where n is the number of sets of
            block layers desired. Each value should be the number of blocks in the
            i-th set.
          block_strides: List of integers representing the desired stride size for
            each of the sets of block layers. Should be same length as block_sizes.
        Raises:
          ValueError: if invalid version is selected.
        """
        super().__init__()

        self._input_shape = (first_in_channels, img_height, img_width)

        is_bottleneck = self.get_is_bottleneck(resnet_size, block_sizes)
        block_class = self.get_block_fn(is_bottleneck)
        block_sizes, block_strides = self.get_blocks(resnet_size, block_sizes, block_strides)

        self.layers = torch.nn.ModuleList()
        self.layers.append(
            Conv2DLayerFixedPadding(
                img_height=img_height,
                img_width=img_width,
                in_channels=first_in_channels,
                out_channels=out_channels,
                kernel_size=kernel_size,
                stride=conv_stride,
            )
        )
        in_channels, img_height, img_width = self.layers[-1].output_shape
        self.layers.append(
            nn.BatchNorm2d(num_features=out_channels, eps=batch_norm_epsilon, momentum=batch_norm_momentum)
        )
        self.layers.append(get_activation("relu"))

        if first_pool_kernel_size:
            self.layers.append(nn.MaxPool2d(kernel_size=first_pool_kernel_size, stride=first_pool_stride, padding=1))
            img_height, img_width = get_img_output_shape(
                img_height=img_height,
                img_width=img_width,
                kernel_size=first_pool_kernel_size,
                stride=first_pool_stride,
                padding=1,
                dilation=1,
            )

        for i, num_blocks in enumerate(block_sizes):
            self.layers.append(
                ResNetBlockLayer(
                    img_height=img_height,
                    img_width=img_width,
                    first_in_channels=in_channels,
                    out_channels=out_channels,
                    is_bottleneck=is_bottleneck,
                    block_fn=block_class,
                    num_blocks=num_blocks,
                    stride=block_strides[i],
                    batch_norm_momentum=batch_norm_momentum,
                    batch_norm_epsilon=batch_norm_epsilon,
                )
            )
            out_channels *= 2
            in_channels, img_height, img_width = self.layers[-1].output_shape

        for layer in self.layers:
            logger.debug(f"   {layer._get_name()}")

        self._output_shape = (in_channels, img_height, img_width)

    def get_is_bottleneck(self, resnet_size: int, block_sizes: List[int]) -> bool:
        if (resnet_size is not None and resnet_size >= 50) or (block_sizes is not None and sum(block_sizes) >= 16):
            return True
        return False

    def get_block_fn(self, is_bottleneck: bool) -> Union[ResNetBlock, ResNetBottleneckBlock]:
        if is_bottleneck:
            return ResNetBottleneckBlock
        return ResNetBlock

    def get_blocks(self, resnet_size: int, block_sizes: List[int], block_strides: List[int]) -> Tuple[List[int]]:
        if block_sizes is None:
            block_sizes = get_resnet_block_sizes(resnet_size)
        if block_strides is None:
            block_strides = [1] + [2 for _ in range(len(block_sizes) - 1)]
        return block_sizes, block_strides

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        hidden = inputs
        for layer in self.layers:
            hidden = layer(hidden)

        return hidden

    @property
    def output_shape(self) -> torch.Size:
        return torch.Size(self._output_shape)

    @property
    def input_shape(self) -> torch.Size:
        return torch.Size(self._input_shape)


################################################################################
# The following code for ResNet is adapted from the TensorFlow implementation
# https://github.com/tensorflow/models/blob/master/official/resnet/resnet_model.py
################################################################################

################################################################################
# Convenience functions for building the ResNet model.
################################################################################
resnet_choices = {
    8: [1, 2, 2],
    14: [1, 2, 2],
    18: [2, 2, 2, 2],
    34: [3, 4, 6, 3],
    50: [3, 4, 6, 3],
    101: [3, 4, 23, 3],
    152: [3, 8, 36, 3],
    200: [3, 24, 36, 3],
}


def get_resnet_block_sizes(resnet_size):
    """Retrieve the size of each block_layer in the ResNet model.

    The number of block layers used for the Resnet model varies according
    to the size of the model. This helper grabs the layer set we want, throwing
    an error if a non-standard size has been selected.
    Args:
      resnet_size: The number of convolutional layers needed in the model.
    Returns:
      A list of block sizes to use in building the model.
    Raises:
      KeyError: if invalid resnet_size is received.
    """
    try:
        return resnet_choices[resnet_size]
    except KeyError:
        err = "Could not find layers for selected Resnet size.\n" "Size received: {}; sizes allowed: {}.".format(
            resnet_size, resnet_choices.keys()
        )
        raise ValueError(err)



# !/usr/bin/env python

from dataclasses import dataclass
from typing import Any, List

from dataclasses_json import dataclass_json

try:
    from ray.tune import ExperimentAnalysis
except ImportError:
    ExperimentAnalysis = Any


@dataclass_json
@dataclass
class TrialResults:
    parameters: dict
    metric_score: float
    training_stats: dict
    eval_stats: dict


@dataclass
class HyperoptResults:
    ordered_trials: List[TrialResults]
    experiment_analysis: ExperimentAnalysis



import ludwig.schema.features.audio_feature
import ludwig.schema.features.bag_feature
import ludwig.schema.features.binary_feature
import ludwig.schema.features.category_feature
import ludwig.schema.features.date_feature
import ludwig.schema.features.h3_feature
import ludwig.schema.features.image_feature
import ludwig.schema.features.number_feature
import ludwig.schema.features.sequence_feature
import ludwig.schema.features.set_feature
import ludwig.schema.features.text_feature
import ludwig.schema.features.timeseries_feature
import ludwig.schema.features.vector_feature  # noqa



"""Public API: Common typing for Ludwig dictionary parameters."""

from typing import Any, Dict

FeatureConfigDict = Dict[str, Any]
"""Dictionary of parameters used to configure an input or output feature.

https://ludwig.ai/latest/configuration/features/supported_data_types/
"""

ModelConfigDict = Dict[str, Any]
"""Dictionary representation of the ModelConfig object.

https://ludwig.ai/latest/configuration/
"""

TrainingSetMetadataDict = Dict[str, Any]
"""Training set metadata, which consists of internal configuration parameters."""

PreprocessingConfigDict = Dict[str, Any]
"""Dictionary of parameters used to configure preprocessing.

May be type-defaults global preprocessing or feature-specific preprocessing.
https://ludwig.ai/latest/configuration/preprocessing/
"""

HyperoptConfigDict = Dict[str, Any]
"""Dictionary of parameters used to configure hyperopt.

https://ludwig.ai/latest/configuration/hyperparameter_optimization/
"""

TrainerConfigDict = Dict[str, Any]
"""Dictionary of parameters used to configure training.

https://ludwig.ai/latest/configuration/trainer/
"""

FeatureTypeDefaultsDict = Dict[str, FeatureConfigDict]
"""Dictionary of type to parameters that configure the defaults for that feature type.

https://ludwig.ai/latest/configuration/defaults/
"""

FeatureMetadataDict = Dict[str, Any]
"""Metadata for a specific feature like idx2str."""

FeaturePostProcessingOutputDict = Dict[str, Any]
"""Output from feature post-processing."""



# Register all preprocessors
from ludwig.schema.features.preprocessing import (  # noqa
    audio,
    bag,
    binary,
    category,
    date,
    h3,
    image,
    number,
    sequence,
    set,
    text,
    timeseries,
    vector,
)



import pytest
import torch

from ludwig.features.audio_feature import AudioInputFeature
from ludwig.schema.features.audio_feature import AudioInputFeatureConfig
from ludwig.schema.utils import load_config_with_kwargs
from tests.integration_tests.utils import audio_feature

BATCH_SIZE = 2
SEQ_SIZE = 20
AUDIO_W_SIZE = 16
DEFAULT_OUTPUT_SIZE = 256


@pytest.mark.parametrize("enc_encoder", ["stacked_cnn", "parallel_cnn", "stacked_parallel_cnn", "rnn", "cnnrnn"])
def test_audio_feature(enc_encoder):
    # synthetic audio tensor
    audio_tensor = torch.randn([BATCH_SIZE, SEQ_SIZE, AUDIO_W_SIZE], dtype=torch.float32)

    # generate audio feature config
    audio_feature_config = audio_feature(
        folder=".", encoder={"type": enc_encoder, "max_sequence_length": SEQ_SIZE, "embedding_size": AUDIO_W_SIZE}
    )

    # instantiate audio input feature object
    audio_feature_config, _ = load_config_with_kwargs(AudioInputFeatureConfig, audio_feature_config)
    audio_input_feature = AudioInputFeature(audio_feature_config)

    # pass synthetic audio tensor through the audio input feature
    encoder_output = audio_input_feature(audio_tensor)

    # confirm correctness of the the audio encoder output
    assert isinstance(encoder_output, dict)
    assert "encoder_output" in encoder_output
    assert isinstance(encoder_output["encoder_output"], torch.Tensor)
    if enc_encoder == "passthrough":
        assert encoder_output["encoder_output"].shape == (BATCH_SIZE, SEQ_SIZE, AUDIO_W_SIZE)
    else:
        assert encoder_output["encoder_output"].shape == (BATCH_SIZE, DEFAULT_OUTPUT_SIZE)



import pytest

from ludwig.modules.mlp_mixer_modules import MixerBlock, MLP, MLPMixer

from .test_utils import assert_output_shapes


@pytest.mark.parametrize("in_features,hidden_size,out_features", [(3, 8, 8), (8, 64, 32)])
def test_mlp(in_features: int, hidden_size: int, out_features: int):
    assert_output_shapes(module=MLP(in_features, hidden_size, out_features), input_shape=(in_features,))


@pytest.mark.parametrize("embed_size,n_patches,token_dim,channel_dim", [(512, 49, 2048, 256)])
def test_mixer_block(
    embed_size: int,
    n_patches: int,
    token_dim: int,
    channel_dim: int,
):
    assert_output_shapes(
        module=MixerBlock(embed_size, n_patches, token_dim, channel_dim), input_shape=(n_patches, embed_size)
    )


@pytest.mark.parametrize("img_height,img_width,in_channels", [(224, 224, 3)])
def test_mlp_mixer(img_height: int, img_width: int, in_channels: int):
    assert_output_shapes(module=MLPMixer(img_height, img_width, in_channels), input_shape=(3, img_height, img_width))



from typing import Any, Dict, List, Optional, Union

from ludwig.api_annotations import DeveloperAPI
from ludwig.schema import common_fields
from ludwig.schema import utils as schema_utils
from ludwig.schema.metadata import COMBINER_METADATA
from ludwig.schema.utils import ludwig_dataclass


@DeveloperAPI
@ludwig_dataclass
class CommonTransformerConfig:
    """Common transformer parameter values."""

    dropout: float = schema_utils.FloatRange(
        default=0.1,
        min=0,
        max=1,
        description="Dropout rate for the transformer block.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["dropout"],
    )

    transformer_output_size: int = schema_utils.NonNegativeInteger(
        default=256,
        description="Size of the fully connected layer after self attention in the transformer block. This is usually "
        "the same as `hidden_size` and `embedding_size`.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["transformer_output_size"],
    )

    hidden_size: int = schema_utils.NonNegativeInteger(
        default=256,
        description="The number of hidden units of the TransformerStack as well as the dimension that each incoming "
        "input feature is projected to before feeding to the TransformerStack.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["hidden_size"],
    )

    num_layers: int = schema_utils.PositiveInteger(
        default=1,
        description="The number of transformer layers.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["num_layers"],
    )

    num_heads: int = schema_utils.NonNegativeInteger(
        default=8,
        description="Number of heads of the self attention in the transformer block.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["num_heads"],
    )

    use_bias: bool = schema_utils.Boolean(
        default=True,
        description="Whether the layer uses a bias vector.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["use_bias"],
    )

    bias_initializer: Union[str, Dict] = common_fields.BiasInitializerField()

    weights_initializer: Union[str, Dict] = common_fields.WeightsInitializerField()

    # TODO(#1673): Add conditional logic for fields like this one:
    num_fc_layers: int = schema_utils.NonNegativeInteger(
        default=0,
        description="The number of stacked fully connected layers (only applies if `reduce_output` is not null).",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["num_fc_layers"],
    )

    output_size: int = schema_utils.PositiveInteger(
        default=256,
        description="Output size of a fully connected layer.",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["output_size"],
    )

    norm: Optional[str] = common_fields.NormField()

    norm_params: Optional[dict] = common_fields.NormParamsField()

    fc_layers: Optional[List[Dict[str, Any]]] = common_fields.FCLayersField()

    fc_dropout: float = common_fields.DropoutField()

    fc_activation: str = schema_utils.ActivationOptions(
        default="relu",
        parameter_metadata=COMBINER_METADATA["TransformerCombiner"]["fc_activation"],
    )

    fc_residual: bool = common_fields.ResidualField()



import os
import re

import numpy as np
import pytest

try:
    import ray as _ray
except ImportError:
    _ray = None

from ludwig.api import LudwigModel
from ludwig.constants import COLUMN, INPUT_FEATURES, MODEL_TYPE, NAME, OUTPUT_FEATURES, TRAINER
from ludwig.error import ConfigValidationError
from tests.integration_tests import synthetic_test_data
from tests.integration_tests.utils import binary_feature
from tests.integration_tests.utils import category_feature as _category_feature
from tests.integration_tests.utils import generate_data, number_feature

BOOSTING_TYPES = ["gbdt", "goss", "dart"]
TREE_LEARNERS = ["serial", "feature", "data", "voting"]


@pytest.fixture(scope="module")
def local_backend():
    return {"type": "local"}


@pytest.fixture(scope="module")
def ray_backend():
    num_workers = 2
    num_cpus_per_worker = 2
    return {
        "type": "ray",
        "processor": {
            "parallelism": 1,
        },
        "trainer": {
            "use_gpu": False,
            "num_workers": num_workers,
            "resources_per_worker": {
                "CPU": num_cpus_per_worker,
                "GPU": 0,
            },
        },
    }


def category_feature(**kwargs):
    encoder = kwargs.get("encoder", {})
    encoder = {**{"type": "passthrough"}, **encoder}
    kwargs["encoder"] = encoder
    return _category_feature(**kwargs)


def _train_and_predict_gbm(input_features, output_features, tmpdir, backend_config, **trainer_config):
    csv_filename = os.path.join(tmpdir, "training.csv")
    dataset_filename = generate_data(input_features, output_features, csv_filename, num_examples=100)

    config = {
        MODEL_TYPE: "gbm",
        INPUT_FEATURES: input_features,
        OUTPUT_FEATURES: output_features,
        # Disable feature filtering to avoid having no features due to small test dataset,
        # see https://stackoverflow.com/a/66405983/5222402
        TRAINER: {"num_boost_round": 2, "feature_pre_filter": False},
    }

    if trainer_config:
        config[TRAINER].update(trainer_config)

    model = LudwigModel(config, backend=backend_config)
    _, _, output_directory = model.train(
        dataset=dataset_filename,
        output_directory=tmpdir,
        skip_save_processed_input=True,
        skip_save_progress=True,
        skip_save_unprocessed_output=True,
        skip_save_log=True,
    )
    model.load(os.path.join(tmpdir, "api_experiment_run", "model"))
    preds, _ = model.predict(dataset=dataset_filename, output_directory=output_directory, split="test")

    return preds, model


def run_test_gbm_binary(tmpdir, backend_config):
    """Test that the GBM model can train and predict a binary variable (binary classification)."""
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_feature = binary_feature()
    output_features = [output_feature]

    preds, _ = _train_and_predict_gbm(input_features, output_features, tmpdir, backend_config)

    prob_col = preds[output_feature["name"] + "_probabilities"]
    if backend_config["type"] == "ray":
        prob_col = prob_col.compute()
    assert len(prob_col.iloc[0]) == 2
    assert prob_col.apply(sum).mean() == pytest.approx(1.0)


def test_local_gbm_binary(tmpdir, local_backend):
    run_test_gbm_binary(tmpdir, local_backend)


@pytest.mark.distributed
def test_ray_gbm_binary(tmpdir, ray_backend, ray_cluster_5cpu):
    run_test_gbm_binary(tmpdir, ray_backend)


def run_test_gbm_non_number_inputs(tmpdir, backend_config):
    """Test that the GBM model can train and predict with non-number inputs."""
    input_features = [binary_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_feature = binary_feature()
    output_features = [output_feature]

    preds, _ = _train_and_predict_gbm(input_features, output_features, tmpdir, backend_config)

    prob_col = preds[output_feature["name"] + "_probabilities"]
    if backend_config["type"] == "ray":
        prob_col = prob_col.compute()
    assert len(prob_col.iloc[0]) == 2
    assert prob_col.apply(sum).mean() == pytest.approx(1.0)


def test_local_gbm_non_number_inputs(tmpdir, local_backend):
    run_test_gbm_non_number_inputs(tmpdir, local_backend)


@pytest.mark.distributed
def test_ray_gbm_non_number_inputs(tmpdir, ray_backend, ray_cluster_5cpu):
    run_test_gbm_non_number_inputs(tmpdir, ray_backend)


def run_test_gbm_category(vocab_size, tmpdir, backend_config):
    """Test that the GBM model can train and predict a categorical output (multiclass classification)."""
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_feature = category_feature(decoder={"vocab_size": vocab_size})
    output_features = [output_feature]

    preds, _ = _train_and_predict_gbm(input_features, output_features, tmpdir, backend_config)

    prob_col = preds[output_feature["name"] + "_probabilities"]
    if backend_config["type"] == "ray":
        prob_col = prob_col.compute()
    assert len(prob_col.iloc[0]) == vocab_size
    assert prob_col.apply(sum).mean() == pytest.approx(1.0)


@pytest.mark.parametrize("vocab_size", [2, 3])
def test_local_gbm_category(vocab_size, tmpdir, local_backend):
    run_test_gbm_category(vocab_size, tmpdir, local_backend)


@pytest.mark.distributed
@pytest.mark.parametrize("vocab_size", [2, 3])
def test_ray_gbm_category(vocab_size, tmpdir, ray_backend, ray_cluster_5cpu):
    run_test_gbm_category(vocab_size, tmpdir, ray_backend)


def run_test_gbm_number(tmpdir, backend_config):
    """Test that the GBM model can train and predict a numerical output (regression)."""
    # Given a dataset with a single input feature and a single output feature,
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_feature = number_feature()
    output_features = [output_feature]

    # When we train a GBM model on the dataset,
    preds, _ = _train_and_predict_gbm(input_features, output_features, tmpdir, backend_config)

    # Then the predictions should be included in the output
    pred_col = preds[output_feature["name"] + "_predictions"]
    if backend_config["type"] == "ray":
        pred_col = pred_col.compute()
    assert pred_col.dtype == float


def test_local_gbm_number(tmpdir, local_backend):
    run_test_gbm_number(tmpdir, local_backend)


@pytest.mark.distributed
def test_ray_gbm_number(tmpdir, ray_backend, ray_cluster_5cpu):
    run_test_gbm_number(tmpdir, ray_backend)


def test_hummingbird_conversion_binary(tmpdir, local_backend):
    """Verify that Hummingbird conversion predictions match LightGBM predictions for binary outputs."""
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_features = [binary_feature()]
    output_feature = f'{output_features[0]["name"]}_probabilities'

    # Train a model and predict using the LightGBM interface
    preds_lgbm, model = _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend)
    probs_lgbm = preds_lgbm[output_feature]

    # Predict using the Hummingbird compiled model
    with model.model.compile():
        preds_hb, _ = model.predict(dataset=os.path.join(tmpdir, "training.csv"), split="test")
        probs_hb = preds_hb[output_feature]

    # sanity check Hummingbird probabilities equal to LightGBM probabilities
    assert np.allclose(np.stack(probs_hb.values), np.stack(probs_lgbm.values), rtol=1e-6, atol=1e-6)


def test_hummingbird_conversion_regression(tmpdir, local_backend):
    """Verify that Hummingbird conversion predictions match LightGBM predictions for numeric outputs."""
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_features = [number_feature()]

    # Train a model and predict using the LightGBM interface
    preds_lgbm, model = _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend)

    # Predict using the Hummingbird compiled model
    with model.model.compile():
        preds_hb, _ = model.predict(dataset=os.path.join(tmpdir, "training.csv"), split="test")

    # sanity check Hummingbird prediction equal to LightGBM prediction
    assert np.allclose(preds_hb, preds_lgbm, rtol=1e-6, atol=1e-6)


@pytest.mark.parametrize("vocab_size", [2, 3])
def test_hummingbird_conversion_category(vocab_size, tmpdir, local_backend):
    """Verify that Hummingbird conversion predictions match LightGBM predictions for categorical outputs."""
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_features = [category_feature(decoder={"vocab_size": vocab_size})]

    # Train a model and predict using the LightGBM interface
    preds_lgbm, model = _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend)
    output_feature = next(iter(model.model.output_features.values()))
    output_feature_name = f"{output_feature.column}_probabilities"
    probs_lgbm = np.stack(preds_lgbm[output_feature_name].to_numpy())

    # Predict using the Hummingbird compiled model
    with model.model.compile():
        preds_hb, _ = model.predict(dataset=os.path.join(tmpdir, "training.csv"), split="test")
        probs_hb = np.stack(preds_hb[output_feature_name].to_numpy())

    # sanity check Hummingbird probabilities equal to LightGBM probabilities
    assert np.allclose(probs_hb, probs_lgbm, rtol=1e-6, atol=1e-6)


def test_loss_decreases(tmpdir, local_backend):
    input_features, output_features = synthetic_test_data.get_feature_configs()

    config = {
        MODEL_TYPE: "gbm",
        "input_features": input_features,
        "output_features": output_features,
        # Disable feature filtering to avoid having no features due to small test dataset,
        # see https://stackoverflow.com/a/66405983/5222402
        TRAINER: {
            "num_boost_round": 2,
            "boosting_rounds_per_checkpoint": 1,
            "feature_pre_filter": False,
        },
    }

    generated_data = synthetic_test_data.get_generated_data_for_optimizer()
    model = LudwigModel(config, backend=local_backend)
    train_stats, _, _ = model.train(
        dataset=generated_data.train_df,
        output_directory=tmpdir,
        skip_save_processed_input=True,
        skip_save_progress=True,
        skip_save_unprocessed_output=True,
        skip_save_log=True,
    )

    # retrieve training losses for first and last entries.
    train_losses = train_stats["training"]["combined"]["loss"]
    last_entry = len(train_losses)

    # ensure train loss for last entry is less than or equal to the first entry.
    assert train_losses[last_entry - 1] <= train_losses[0]


def test_save_load(tmpdir, local_backend):
    input_features = [number_feature(), category_feature(encoder={"reduce_output": "sum"})]
    output_features = [binary_feature()]

    init_preds, model = _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend)

    # save model
    model.save(tmpdir)

    # load model
    model = LudwigModel.load(tmpdir)
    preds, _ = model.predict(dataset=os.path.join(tmpdir, "training.csv"), split="test")

    assert init_preds.equals(preds)


def test_lgbm_dataset_setup(tmpdir, local_backend):
    """Test that LGBM dataset column name errors are handled."""
    input_features = [number_feature()]
    output_features = [binary_feature()]

    # Overwrite the automatically generated feature/column name with an invalid string.
    input_features[0][NAME] = ",Unnamed: 0"
    input_features[0][COLUMN] = input_features[0][NAME]

    # Test that the custom error is raised (lightgbm.basic.LightGBMError -> ValueError)
    with pytest.raises(ValueError):
        try:
            _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend)
        except ValueError as e:
            # Check that the intended ValueError was raised
            assert re.search("Some column names in the training set contain invalid characters", str(e))
            raise


def test_boosting_type_rf_invalid(tmpdir, local_backend):
    """Test that the Random Forest boosting type is not supported.

    LightGBM does not support model checkpointing for `boosting_type=rf`. This test ensures that a schema validation
    error is raised when trying to use random forests.
    """
    input_features = [number_feature()]
    output_features = [binary_feature()]

    with pytest.raises(ConfigValidationError):
        _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend, boosting_type="rf")


@pytest.mark.skip(reason="LightGBMError: Number of class for initial score error")
def test_goss_deactivate_bagging(tmpdir, local_backend):
    """Test that bagging is disabled for the GOSS boosting type.

    TODO: Re-enable when GOSS is supported: https://github.com/ludwig-ai/ludwig/issues/2988
    """
    input_features = [number_feature()]
    output_features = [binary_feature()]

    _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend, boosting_type="goss", bagging_freq=5)


@pytest.mark.parametrize("tree_learner", TREE_LEARNERS)
def test_boosting_type_null_invalid(tree_learner, tmpdir, local_backend):
    """Test that the null boosting type is disabled.

    `boosting_type: null` defaults to "gbdt", and it was removed to avoid confusing GBM trainer settings.
    """
    input_features = [number_feature()]
    output_features = [binary_feature()]

    with pytest.raises(ConfigValidationError):
        _train_and_predict_gbm(
            input_features, output_features, tmpdir, local_backend, boosting_type=None, tree_learner=tree_learner
        )


@pytest.mark.parametrize("boosting_type", BOOSTING_TYPES)
def test_tree_learner_null_invalid(boosting_type, tmpdir, local_backend):
    """Test that the null tree learner is disabled.

    `tree_learner: null` defaults to "serial", and it was removed to avoid confusing GBM trainer settings.
    """
    input_features = [number_feature()]
    output_features = [binary_feature()]

    with pytest.raises(ConfigValidationError):
        _train_and_predict_gbm(
            input_features, output_features, tmpdir, local_backend, boosting_type=boosting_type, tree_learner=None
        )


def test_dart_boosting_type(tmpdir, local_backend):
    """Test that DART does not error during eval due to progress tracking."""
    input_features = [number_feature()]
    output_features = [binary_feature()]

    _train_and_predict_gbm(input_features, output_features, tmpdir, local_backend, boosting_type="dart")
